/*
Bullet Continuous Collision Detection and Physics Library
Copyright (c) 2003-2006 Erwin Coumans  http://continuousphysics.com/Bullet/

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it freely,
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/

#include "btBulletDynamicsCommon.h"
#include "LinearMath/btIDebugDraw.h"
#include "GLDebugDrawer.h"
#include "Hinge2VehicleDemo.h"
#include "btWheelShapes.h"
#include "GL_ShapeDrawer.h"
#include "GlutStuff.h"
#include "btTriangleMeshEx.h"
#include "btHACDCompoundShape.h"
//#include "btRaycastVehicleEx.h"


//#define USE_PARALLEL_DISPATCHER 1
#ifdef USE_PARALLEL_DISPATCHER
#include "../../Extras/BulletMultiThreaded/SpuGatheringCollisionDispatcher.h"
#include "../../Extras/BulletMultiThreaded/Win32ThreadSupport.h"
#include "../../Extras/BulletMultiThreaded/SpuNarrowPhaseCollisionTask/SpuGatheringCollisionTask.h"
#endif//USE_PARALLEL_DISPATCHER

#include <BulletWorldImporter/btBulletWorldImporter.h>
#include "GLDebugDrawer.h"

 //region Twickable Definitions
 #define BETTER_STABILITY_OF_WHEEL_TRAIN_SHAPES 1 //2 // through a CustomMaterialCallback btWheelTrainShapes are forced to have 0 friction on their conic subparts (this makes the train stop if falling outside the tracks)
                                                    // setting this to 2: the same but for locomotive wheels only (less stable, but better performance, since the CF_CUSTOM_MATERIAL_CALLBACK flag is set only for locomotive wheels)
//#define USE_ANISOTROPIC_FRICTION_FOR_WHEELS		1	// Due to the lack of differential, this makes easier for the vehicles to fall on their side when steering...
//#define USE_LINEAR_FRICTION_MODEL 	1		// defines USE_CUSTOM_MATERIAL_CALLBACK
//#define USE_LINEAR_RESTITUTION_MODEL	1		// defines USE_CUSTOM_MATERIAL_CALLBACK
//endregion

 //region Twickable Globals
const btScalar g_xAnisotropicFrictionFactorForWheels(1.5f);	// If the USE_ANISOTROPIC_FRICTION_FOR_WHEELS is defined friction.x = g_xAnisotropicFrictionFactorForWheels*friction for all wheel shapes except train wheels
const btScalar g_MAX_FRICTION  = 100.f;//10.f; 		// Used ONLY when USE_CUSTOM_MATERIAL_CALLBACK is used (I don't know what happens if friction>1, to be honest...)
const btScalar g_MAX_RESTITUTION  = 100.f;//2.f; 	// Used ONLY when USE_CUSTOM_MATERIAL_CALLBACK is used (I don't know what happens if restitution>1, to be honest...)
//endregion


//--This stuff is used by the forklift constraint only------
struct btSliderHelperStruct	{
void lock() const	{
	if (!m_forkSlider) return;
	const btScalar linDepth = m_forkSlider->getLinearPos();
	m_forkSlider->setPoweredLinMotor(false);
	if(linDepth <= m_linearLowerLimit)	{
		m_forkSlider->setLowerLinLimit(m_linearLowerLimit);
		m_forkSlider->setUpperLinLimit(m_linearLowerLimit);
	}
	else if(linDepth > m_linearUpperLimit)	{
		m_forkSlider->setLowerLinLimit(m_linearUpperLimit);
		m_forkSlider->setUpperLinLimit(m_linearUpperLimit);
	}
	else	{
		m_forkSlider->setLowerLinLimit(linDepth);
		m_forkSlider->setUpperLinLimit(linDepth);
	}
}
void slideUp() const 	{
	m_forkSlider->setLowerLinLimit(m_linearLowerLimit);
	m_forkSlider->setUpperLinLimit(m_linearUpperLimit);
	m_forkSlider->setPoweredLinMotor(true);
	m_forkSlider->setMaxLinMotorForce(m_linearMotorMaxForce);
	m_forkSlider->setTargetLinMotorVelocity(m_linearMotorTargetVelocity);
}
void slideDown() const	{
	m_forkSlider->setLowerLinLimit(m_linearLowerLimit);
	m_forkSlider->setUpperLinLimit(m_linearUpperLimit);
	m_forkSlider->setPoweredLinMotor(true);
	m_forkSlider->setMaxLinMotorForce(m_linearMotorMaxForce);
	m_forkSlider->setTargetLinMotorVelocity(-m_linearMotorTargetVelocity);
}
	btSliderHelperStruct(btSliderConstraint* forkSlider,btScalar linearLowerLimit,btScalar linearUpperLimit,btScalar linearMotorMaxForce=btScalar(10),btScalar linearMotorTargetVelocity=btScalar(1))
	: m_forkSlider(forkSlider), m_linearLowerLimit(linearLowerLimit),m_linearUpperLimit(linearUpperLimit),m_linearMotorMaxForce(linearMotorMaxForce),m_linearMotorTargetVelocity(linearMotorTargetVelocity)
	{}

	btSliderConstraint* m_forkSlider;
	btScalar m_linearLowerLimit;
	btScalar m_linearUpperLimit;
	btScalar m_linearMotorMaxForce;
	btScalar m_linearMotorTargetVelocity;
};
static btAlignedObjectArray <btSliderHelperStruct*> g_sliderHelperStructs;
//----------------------------------------------------------
Hinge2VehicleDemo* g_vehicleDemo = NULL;	//Me
void g_internalTickCallback(btDynamicsWorld *world, btScalar timeStep)	{
	if (g_vehicleDemo) g_vehicleDemo->preTickCallback(timeStep);
}
typedef void (*btInternalTickCallback)(btDynamicsWorld *world, btScalar timeStep);
// We get pressed keys in glut special key event handler, but we use (internal) impulses that need to be applied inside
// the tick callback: so we must store our input somehow... (it's not just a question of querying glut for pressed keys,
// we need to buffer them).
bool g_vehicleKeyPressedLeft(false);
bool g_vehicleKeyPressedRight(false);
bool g_vehicleKeyPressedUp(false);
bool g_vehicleKeyPressedDown(false);
bool g_vehicleKeyPressedSpecialActionUp(false);
bool g_vehicleKeyPressedSpecialActionDown(false);



 //region Custom Material Callback Stuff
inline static void ClampInPlace(btScalar& value,btScalar minValue,btScalar maxValue)	{
	if (value < minValue) value = minValue;
	else if (value > maxValue) value = maxValue;
}
inline static btScalar CalculateCombinedFriction(btScalar friction0,btScalar friction1)	{
	btScalar friction =
						#ifndef USE_LINEAR_FRICTION_MODEL
						friction0 * friction1;	// Default behavior AFAIK
						#else //USE_LINEAR_FRICTION_MODEL
						(friction0 + friction1)*0.5f;
						#endif //USE_LINEAR_FRICTION_MODEL
	ClampInPlace(friction,0,g_MAX_FRICTION);	// Does negative friction have any sense ?
	return friction;
}
inline static btScalar CalculateCombinedRestitution(btScalar restitution0,btScalar restitution1)	{
	btScalar restitution =
						   #ifndef USE_LINEAR_RESTITUTION_MODEL
						   restitution0 * restitution1;	// Default behavior AFAIK
						   #else //USE_LINEAR_RESTITUTION_MODEL
						   (restitution0 + restitution1)*0.5f;
						   #endif //USE_LINEAR_RESTITUTION_MODEL

	ClampInPlace(restitution,0,g_MAX_RESTITUTION);// Does negative restitution have any sense ?
	return restitution;
}

//#define USE_AN_OLDER_BULLET_LIBRARY_VERSION //Tweakable in case of compilation errors
#ifndef USE_AN_OLDER_BULLET_LIBRARY_VERSION
static bool CustomMaterialCombinerCallback(btManifoldPoint& cp,	const btCollisionObjectWrapper* colObj0Wrapper,int partId0,int index0,const btCollisionObjectWrapper* colObj1Wrapper,int partId1,int index1)
#else //USE_AN_OLDER_BULLET_LIBRARY_VERSION
static bool CustomMaterialCombinerCallback(btManifoldPoint& cp,	const btCollisionObject* colObj0,int partId0,int index0,const btCollisionObject* colObj1,int partId1,int index1)
#endif //USE_AN_OLDER_BULLET_LIBRARY_VERSION
{
#ifndef USE_AN_OLDER_BULLET_LIBRARY_VERSION
    const btCollisionObject* colObj0 = colObj0Wrapper->getCollisionObject();
    const btCollisionObject* colObj1 = colObj1Wrapper->getCollisionObject();
    const btCollisionShape* colShape0 = colObj0->getCollisionShape();
    const btCollisionShape* colShape1 = colObj1->getCollisionShape();
#else //USE_AN_OLDER_BULLET_LIBRARY_VERSION
    const btCollisionShape* colShape0 = colObj0->getRootCollisionShape();
    const btCollisionShape* colShape1 = colObj1->getRootCollisionShape();
#endif//USE_AN_OLDER_BULLET_LIBRARY_VERSION

    #ifdef BETTER_STABILITY_OF_WHEEL_TRAIN_SHAPES
    const btWheelTrainShape* ts0 = btWheelTrainShape::upcast(colShape0);
    if (ts0) {
        //printf("partId0 = %d index0 = %d\n",partId0,index0);
        if (index0!=0) {cp.m_combinedFriction = cp.m_combinedRestitution = 0;return true;}

        // Here we are colliding with the cylindrical part of the wheel shape,
        // but we don't want to allow the optional USE_LINEAR_FRICTION_MODEL/USE_LINEAR_RESTITUTION_MODEL definitions to kick in:
        btScalar friction = colObj0->getFriction()*colObj1->getFriction();	// Default behavior
        ClampInPlace(friction,0,g_MAX_FRICTION);
        cp.m_combinedFriction = friction;
        cp.m_combinedRestitution = 0;
        return true;
    }
    const btWheelTrainShape* ts1 = btWheelTrainShape::upcast(colShape1);
    if (ts1) {
        //printf("partId1 = %d index1 = %d\n",partId1,index1);
        if (index1!=0) {cp.m_combinedFriction = cp.m_combinedRestitution = 0;return true;}

        // Here we are colliding with the cylindrical part of the wheel shape,
        // but we don't want to allow the optional USE_LINEAR_FRICTION_MODEL/USE_LINEAR_RESTITUTION_MODEL definitions to kick in:
        btScalar friction = colObj0->getFriction()*colObj1->getFriction();	// Default behavior
        ClampInPlace(friction,0,g_MAX_FRICTION);
        cp.m_combinedFriction = friction;
        cp.m_combinedRestitution = 0;
        return true;
    }
    #endif //BETTER_STABILITY_OF_WHEEL_TRAIN_SHAPES


	// Other optional stuff can be placed here...

    //-------------------------------------------

    cp.m_combinedFriction = CalculateCombinedFriction(colObj0->getFriction(),colObj1->getFriction());
    cp.m_combinedRestitution = CalculateCombinedRestitution(colObj0->getRestitution(),colObj1->getRestitution());
	//this return value is currently ignored, but to be on the safe side: return false if you don't calculate friction
        return true;
}
extern ContactAddedCallback	gContactAddedCallback;
//endregion

 //region A Few Static Helper Methods
/*
SIMD_FORCE_INLINE static const btVector3& GetCOMOffset(const btRigidBody& RB)	{
	return static_cast < const btDefaultMotionState* > (RB.getMotionState())->m_centerOfMassOffset.getOrigin();
}
*/
//This static helper method was intended to remove the COM offset from the rigid body world transform (it's called by updateCamera() for camera targeting purposes), but it does not work, since the COM is still present...
SIMD_FORCE_INLINE static const btTransform GetGraphicWorldTransform(const btRigidBody& RB) {
	#ifdef USE_COLLISION_OBJECT_USER_POINTER
	const btMyDefaultMotionState* ms;
	if (ms = btMyDefaultMotionState::upcast(&RB)) {
		btTransform T;
		ms->getWorldTransformOutsideCOM(T);
		return T;
	}
	#endif //USE_COLLISION_OBJECT_USER_POINTER
	return static_cast < const btDefaultMotionState* > (RB.getMotionState())->m_graphicsWorldTrans;
}
// This is instead a very useful method IMO, but I don't know how fast it is. It's called by updateCamera()
SIMD_FORCE_INLINE static btVector3 GetAABBExtents(const btCollisionShape* shape)	{
		if (shape)	{
			btVector3 aabbMin,aabbMax;
			shape->getAabb(btTransform::getIdentity(),aabbMin,aabbMax);
			return (aabbMax - aabbMin);
		}
		return btVector3(0,0,0);
}
SIMD_FORCE_INLINE static btVector3 GetAABBExtents(const btRigidBody& RB)	{
		return GetAABBExtents(RB.getCollisionShape());
}
//endregion



Hinge2VehicleDemo::Hinge2VehicleDemo() :
m_activeVehicleIndex(0),
m_minCameraDistance(1.f)
{
	//IDrivingInterface::m_inputSpeed = m_inputSpeed;	//We need to keep them in sync
	//#ifdef NEVER
	m_debugMode |=
					btIDebugDraw::DBG_NoHelpText	// Press h to retrieve info
					// Due to a bug all the debug flags (below) seem to work only after 'h' is pressed at runtime:
					/*
					|
					btIDebugDraw::DBG_DrawWireframe
					| btIDebugDraw::DBG_DrawAabb
					| btIDebugDraw::DBG_DrawConstraints
					| btIDebugDraw::DBG_DrawConstraintLimits
					*/
					;
	//#endif
    //printf("\n\nKEYS:\n=====\n\nCAMERA:\n-------\nHOME - change camera view\nPAGE_UP / PAGE_DOWN - change camera distance\n\nVEHICLES:\n---------\nEND - change vehicle\nARROW KEYS - drive vehicle\n\nGLOBAL INPUT SPEED:\n--------------------\nF1 / F2 - decrease / increase input speed\n\nAll keys must be pressed while the openGL window has focus.\n\n");

}

void Hinge2VehicleDemo::createEmptyDynamicsWorld(bool softBodySupport)	{
	GlutDemoApplicationEx::createEmptyDynamicsWorld(softBodySupport);

	g_vehicleDemo = this;
	if (m_dynamicsWorld) m_dynamicsWorld->setInternalTickCallback(g_internalTickCallback);

    //if (m_dynamicsWorld) m_dynamicsWorld->getSolverInfo().m_splitImpulse = true;	// Optional...
}

void Hinge2VehicleDemo::createGround()	{
	 //region Create Ground

	const btScalar groundHalfDimension(1000);
	const btScalar groundHalfDepth(20);
	btCollisionShape* groundShape = new btBoxShape(btVector3(groundHalfDimension,groundHalfDepth,groundHalfDimension));
    const GLuint groundTexture = TextureHelper::Get("./Media/ground.jpg");
	if (groundTexture>0)	{
			btTriangleMeshEx mesh;
			const btScalar texRep(groundHalfDimension*btScalar(0.25));
			mesh.loadDefaultPlaneMesh(2,groundHalfDimension,groundHalfDimension,texRep,texRep);//,0,0,true);
			Material* mat = mesh.getMeshSubPartMaterial(0);
			if (mat) {
				// We modify it a bit...
				//mat->amb[0]=mat->amb[1]=mat->amb[2]=0.4f;
				//mat->dif[0]=mat->dif[1]=mat->dif[2]=1.0f;
				mat->spe[0]=mat->spe[1]=mat->spe[2]=0.0f;
				mat->shi = 0.0f;
			}
                        //mesh.setMeshSubPartTextureId(0,groundTexture);	// Here we set our groud texture (or we can set it as a mesh override later).

                        this->AssignMesh(groundShape,mesh,true);    // last arg = use flat normals
						GetMeshOverrides(*groundShape).offset = btVector3(0,groundHalfDepth,0);	// This is because we are assigning a planar mesh to a cube
            			GetMeshOverrides(*groundShape).textureArray.push_back(groundTexture);   // (or we can directly set it to 'mesh' 3 lines above. This approach leaves the original mesh without textures assigned, so that shapes can have no texture if needed).
                        //mesh.displayDebugInfo();
			//printf("GROUND SHOULD BE OK\n");
	}

	btTransform	startTransform = btTransform::getIdentity();
	startTransform.setOrigin(btVector3(0.f,-20,0.f));	// so we are sure that groundLevel = 0 is above ground.


	float mass = 0.f;
    btRigidBody* staticBody = localCreateRigidBody(mass, startTransform,groundShape);
	staticBody->setFriction(0.75f);	// probably overridden if custom material callback is enabled...

    staticBody->setCollisionFlags(staticBody->getCollisionFlags() | btCollisionObject::CF_STATIC_OBJECT);////btCollisionObject::CF_KINEMATIC_OBJECT);//CF_STATIC_OBJECT);
	//enable custom material callback
	#ifdef USE_CUSTOM_MATERIAL_CALLBACK
	staticBody->setCollisionFlags(staticBody->getCollisionFlags()  | btCollisionObject::CF_CUSTOM_MATERIAL_CALLBACK);
	#endif //USE_CUSTOM_MATERIAL_CALLBACK
//endregion

}


void Hinge2VehicleDemo::initPhysics(){
    myinit();
	setTexturing(true);
    setShadows(true);	//false = Faster
    g_colorMode = 1;
    g_renderMode = 2;

    gContactAddedCallback = CustomMaterialCombinerCallback;

	createEmptyDynamicsWorld();
	createGround();

	clientResetScene();

    initPhysicsHinge2Vehicles();

	// Needed to setup camera system (at the end of initPhysics()):
	refreshCameraVehicleOffset();
	setCameraDistance(10);
	m_ele=22.5;

    printf("\n----------------\nGENERAL KEYS:\n----------------\nUse 'arrow keys' to drive vehicles (use one tap in the opposite direction to brake).\nUse SHIFT + 'arrow keys' to adjust camera.\nPress 'PgUp/PgDw'' or use 'mouse wheel'' to zoom.\nPress 'F4' to change vehicle.\nPress 'F7/F8'' to operate the forklift.\nPress 'SPACE' to reset the simulation.\nPress 'r' to see the collision shapes.\nPress 'g' to toggle shadows on and off.\nPress 'o' to toggle ortho mode.\nPress 'R (uppercase)' to show/hide raycast screenshot.\nUse RMB to grab and MMB to shoot a box.\nPress 'h' and 'w' to see debug stuff.\nPress 'q' to quit.\nAnd please do it while the OpenGL window has focus.\n\n");

}

void Hinge2VehicleDemo::initPhysicsHinge2Vehicles(){
	btHACDCompoundShape::Params params;

	btTriangleMeshEx vehicleWheelMesh;
    if (vehicleWheelMesh.loadFromOBJFile("./Media/wheel/wheel.obj"))  {
    	vehicleWheelMesh.center();
	}
    else printf("Error. Can't load file \"wheel.obj\"\n");


    #define TEST_SINGLE_VEHICLE
	#ifdef TEST_SINGLE_VEHICLE
	// Build Simple vehicle
	{

	// I suggest playing around with params here, before doing it with the more complex vehicles that follows...
	const btScalar chassisHeight(1.5f);
	const btScalar wheelHeight(chassisHeight*0.5f);
	const btTransform chassisTransform(btQuaternion::getIdentity(),btVector3(0.f,chassisHeight,0.f));
    const btVector3 chassisCOMOffset(0.f,-(chassisHeight*0.5f)*0.3f,0.f);	// center of mass offset
 	btTriangleMeshEx mesh;
    if (mesh.loadFromOBJFile("./Media/car/car.obj"))  {
       	mesh.center();
        mesh.setHeight(chassisHeight);
		vehicleWheelMesh.setHeight(wheelHeight);
		params.nClusters = 2;
		params.addFacesPoints = true;

        // Big stuff, I know, but note that almost all the params are in their default state...
		btHinge2Vehicle* vehicle = localCreateHinge2Vehicle(
											 chassisTransform,//const btTransform& startTransform,
											 4,//const unsigned short numberOfWheels = 4,
											 2,//const unsigned short numberOfSteeringWheels = 2,
											 0,//const unsigned short tractionMode = 0,			// 0 = steering wheels, 1 = all the other wheels, 2 = all wheels ( NO DIFFERENTIAL IN ALL CASES )
											 0,//const unsigned short wheelShapeTypeIndex = 2,	// 0 = cylinder, 1 = scaled sphere, 2 = very complex wheel shape, 3 = train wheel shape

											 0,//const unsigned short wheelPositionMode = 0,	// 0 = NORMAL, 1 = PUT ALL WHEELS AT THE FRONT, 2 = PUT ALL WHEELS AT THE BACK

											 mesh,//const btTriangleMeshEx& chassisCenteredMesh,
											 vehicleWheelMesh,//const btTriangleMeshEx& wheelCenteredMesh,

											 params,//const btHACDCompoundShape::Params& chassisHACDCompoundShapeParams=btHACDCompoundShape::Params(),
											 true,//const bool useFlatNormalsForChassisMesh= false,

											 btScalar(-1.f),//const btScalar _chassisMass=btScalar(-1.f),
											 btScalar(-1.f),//const btScalar _wheelMass=btScalar(-1),

											 chassisCOMOffset,//const btVector3& chassisCOM=btVector3(0.f,0.f,0.f)
											 btVector3(0.f,0.f,0.f),//btVector3(chassisLocalAABB.x()*0.55,-1.5*chassisLocalAABB.y(),chassisLocalAABB.z()*0.4),//btVector3(0.f,0.f,0.f),//const btVector3& _chassisWheelOffset=btVector3(0.f,0.f,0.f),	// if != 0, you'd better set it using fraction of chassisLocalAABB components to position one of the most distant wheels (+X,-Y,+Z) of the vehicle (// position of the front-right wheel relative to the chassis mesh center)
				 							 btVector3(0.f,0.f,0.f),//const btVector3& _wheelsAsymmetricOffset=btVector3(0.f,0.f,0.f)	// mainly useful to shift all the wheels together to the front/back

											 btScalar(0.5f),//const btScalar chassisFriction=btScalar(0.5f),
											 btScalar(0.0f),//const btScalar chassisRestitution=btScalar(0.0f),
											 btScalar(1.0f),//const btScalar wheelFriction=btScalar(1.0f),
											 btScalar(0.0f),//const btScalar wheelRestitution=btScalar(0.0f),

                                             NULL,//btCollisionShape* reuseThisCollisionShapeForChassis,	//If used, all other params must be set as if it was NULL.  Use it only with chassis shapes previously created by this same method.
											 NULL,//btCollisionShape* reuseThisCollisionShapeForWheel,		//If used, all other params must be set as if it was NULL.  Use it only with wheel shapes previously created by this same method.

											 -1,//const short int chassisCollisionFilterGroup=-1,
											 -1,//const short int chassisCollisionFilterMask=-1,
											 -1,//const short int wheelsCollisionFilterGroup=-1,
											 -1,//const short int wheelsCollisionFilterMask=-1,

											 // These should be left as default unless further tweaking is required...
											 btScalar(-1),//const btScalar verticalSlidersExtension=btScalar(-1),
											 btScalar(-1),//const btScalar verticalSlidersMotorVelocity=btScalar(-1),
											 btScalar(-1)//const btScalar verticalSlidersMotorMaxForce=btScalar(-1)
											 );

		if (vehicle) {
			m_driving_vehicles.push_back(vehicle); // enables driving input and camera targeting
            vehicle->onBrake();		// starts with brakes on (it won't go around by itself)
		}
	}
    else printf("Error. Can't load file \"car.obj\"\n");
	}
	#endif // TEST_SINGLE_VEHICLE


    #define TEST_TRUCK
	#ifdef TEST_TRUCK
	// Build truck
	{
	// Truck engine data
	btHinge2Vehicle* vehicle = NULL;
    const btScalar chassisHeight(3);
	const btTransform chassisT(btQuaternion::getIdentity(),btVector3(20.f,chassisHeight,0.f));

	//const btVector3 chassisWheelOffset(chassisLocalAABB.x()*0.5f-wheelWidth*0.5f,-chassisLocalAABB.y()*0.75f,chassisLocalAABB.z()*0.5f-2.0*wheelRadius);
	const btScalar chassisMass = 750.0f;//-1.f;//200.0f;//-1.f;
	const unsigned short wheelShapeNumber = 2;	// 0 is the default choice (cylinder), 2 should be the most complex wheel shape (but use it for testing it for now...).
	const btScalar wheelMass = 37.5f;//-1.f;//10.0f;//-1.f;

	btVector3 chassisLocalAABB;
	btVector3 chassisCOM;
	btVector3 wheelLocalAabb;
        // Truck engine (driving vehicle)
        {
            btTriangleMeshEx mesh;
            if (mesh.loadFromOBJFile("./Media/truck/engine/truckEngine.obj"))  {
                mesh.center();
                mesh.setHeight(chassisHeight);
                mesh.rotate(btQuaternion(btVector3(0,1,0),SIMD_PI));   // (The model was looking into the -Z direction)
                chassisLocalAABB = mesh.getAabbFullExtents();

     			chassisCOM = btVector3(0.f,-(chassisHeight*0.5f)*0.5f,chassisLocalAABB.z()*0.25f);
				const btScalar wheelWidth(chassisLocalAABB.x()*0.2f);
        		vehicleWheelMesh.setWidth(wheelWidth);
                wheelLocalAabb = vehicleWheelMesh.getAabbFullExtents();

                params.nClusters = 2;
                params.addFacesPoints = false;

				// Truck engine (driving vehicle)
				vehicle = localCreateHinge2Vehicle(
											 chassisT,//const btTransform& startTransform,
											 6,//const unsigned short numberOfWheels = 4,
                                                                                         2,//const unsigned short numberOfSteeringWheels = 2,
											 1,//const unsigned short tractionMode = 0,			// 0 = steering wheels, 1 = not steering wheels, 2 = all wheels ( NO DIFFERENTIAL IN ALL CASES )
											 wheelShapeNumber,//const unsigned short wheelShapeTypeIndex = 2,	// 0 = cyl, 1 = scaled sphere, 2 = very complex wheel shape, 3 = train wheel shape

											 0,//const unsigned short wheelPositionMode = 0,	// 0 = NORMAL, 1 = PUT ALL WHEELS AT THE FRONT, 2 = PUT ALL WHEELS AT THE BACK


											 mesh,//const btTriangleMeshEx& chassisCenteredMesh,
											 vehicleWheelMesh,//const btTriangleMeshEx& wheelCenteredMesh,

											 params,//const btHACDCompoundShape::Params& chassisHACDCompoundShapeParams=btHACDCompoundShape::Params(),
											 false,//const bool useFlatNormalsForChassisMesh= false,

											 chassisMass,//btScalar(-1.f),//const btScalar _chassisMass=btScalar(-1.f),
											 wheelMass,//btScalar(-1.f),//const btScalar _wheelMass=btScalar(-1),

											 chassisCOM,//const btVector3& chassisCOM=btVector3(0.f,0.f,0.f)
                                             btVector3(chassisLocalAABB.x()*0.45f,-chassisLocalAABB.y()*0.55f,0.4f*chassisLocalAABB.z()),//const btVector3& _chassisWheelOffset=btVector3(0.f,0.f,0.f),	// if != 0, you'd better set it using fraction of chassisLocalAABB components to position one of the most distant wheels of the vehicle
				 							 btVector3(0.f,0.f,0.f),//const btVector3& _wheelsAsymmetricOffset=btVector3(0.f,0.f,0.f)	// mainly useful to shift all the wheels together to the front/back

											 btScalar(0.5f),//const btScalar chassisFriction=btScalar(0.5f),
											 btScalar(0.0f),//const btScalar chassisRestitution=btScalar(0.0f),
                                             btScalar(2.0f),//const btScalar wheelFriction=btScalar(1.0f),
											 btScalar(0.0f),//const btScalar wheelRestitution=btScalar(0.0f),

                                             NULL,//btCollisionShape* reuseThisCollisionShapeForChassis,	//If used, all other params must be set as if it was NULL
											 NULL,//btCollisionShape* reuseThisCollisionShapeForWheel,		//If used, all other params must be set as if it was NULL

											 -1,//const short int chassisCollisionFilterGroup=-1,
											 -1,//const short int chassisCollisionFilterMask=-1,
											 -1,//const short int wheelsCollisionFilterGroup=-1,
											 -1,//const short int wheelsCollisionFilterMask=-1,

											 // These should be left as default unless further tweaking is required...
											 btScalar(-1),//const btScalar verticalSlidersExtension=btScalar(-1),
											 btScalar(-1),//const btScalar verticalSlidersMotorVelocity=btScalar(-1),
											 btScalar(-1)//const btScalar verticalSlidersMotorMaxForce=btScalar(-1)
											);
				if (vehicle) {
                    //-------------------------------------------------------------------------------------------
					m_driving_vehicles.push_back(vehicle); // enables driving input and camera targeting
					//This should be the angular velocity of the power wheels (ie: with bigger wheel radius, the linear velocity will be bigger with the same angular velocity):
                    vehicle->setMaxPossibleVelocity(27.0f);		//35.0 is the default
					vehicle->setMinPossibleVelocity(-10.f);
                    vehicle->onBrake();		// starts with brakes on (it won't go around by itself)
				}


            }
            else printf("Error. I can't load \"truckEngine.obj\"\n");
        }

	// Truck trailer data
        btHinge2Vehicle* trailer = NULL;
        btVector3 trailerLocalAABB;//(chassisLocalAABB.x(),chassisLocalAABB.y()*5.0f/btScalar(7),chassisLocalAABB.z()*2.25f);
		const btScalar intersectionLength = chassisLocalAABB.z()*0.15f; // the part of the vehicle length in which the trailer intersect the truck engine vehicle
		btTransform trailerT;
		btVector3 trailerCOM;
        btVector3 trailerWheelOffset;
        const btScalar trailerMass= 300.0f;//400.0f;//-1.f;
	// Note: We can set the 3 masses manually. We can set them to -1 to have them automatically calculated (based on the AABBs). But by doing so, the steering seems to be a a bit limited,
	// mainly because the trailer weight is too heavy (in fact only assigning manually the trailer weight seems to be good as well). Feel free to try different solutions...

        // Truck trailer (passive vehicle)
        {
            btTriangleMeshEx mesh;
            if (mesh.loadFromOBJFile("./Media/truck/trailer/truckTrailer.obj"))  {
                mesh.center();
                mesh.rotate(btQuaternion(btVector3(0,1,0),SIMD_PI));   // (The model was looking into the -Z direction)
                mesh.setWidth(chassisLocalAABB.x());
                trailerLocalAABB = mesh.getAabbFullExtents();
                params.nClusters = 2;
                params.addFacesPoints = false;

                trailerT = btTransform(btQuaternion::getIdentity(),btVector3(20.f,5.0f,-chassisLocalAABB.z()*0.5f-trailerLocalAABB.z()*0.5f+intersectionLength));
				trailerCOM.setValue(0.f,-trailerLocalAABB.y()*0.35f,-trailerLocalAABB.z()*0.25f);
        		trailerWheelOffset.setValue(trailerLocalAABB.x()*0.45f,-trailerLocalAABB.y()*0.45f-chassisLocalAABB.y()*1.25f/btScalar(8),trailerLocalAABB.z()*0.45f);	// We'll use the wheelPositionMode=2, i.e. all wheels will be placed at the back of the vehicle

 				trailer = localCreateHinge2Vehicle(
											 trailerT,//const btTransform& startTransform,
											 4,//const unsigned short numberOfWheels = 4,
											 0,//const unsigned short numberOfSteeringWheels = 2,
											 0,//const unsigned short tractionMode = 0,			// 0 = steering wheels, 1 = not steering wheels, 2 = all wheels ( NO DIFFERENTIAL IN ALL CASES )
											 wheelShapeNumber,//const unsigned short wheelShapeTypeIndex = 2,	// 0 = cyl, 1 = scaled sphere, 2 = very complex wheel shape, 3 = train wheel shape

											 2,//const unsigned short wheelPositionMode = 0,	// 0 = NORMAL, 1 = PUT ALL WHEELS AT THE FRONT, 2 = PUT ALL WHEELS AT THE BACK

											 mesh,//const btTriangleMeshEx& chassisCenteredMesh,
											 vehicleWheelMesh,//const btTriangleMeshEx& wheelCenteredMesh,

											 params,//const btHACDCompoundShape::Params& chassisHACDCompoundShapeParams=btHACDCompoundShape::Params(),
											 true,//const bool useFlatNormalsForChassisMesh= false,

											 trailerMass,//btScalar(-1.f),//const btScalar _chassisMass=btScalar(-1.f),
											 wheelMass,//btScalar(-1.f),//const btScalar _wheelMass=btScalar(-1),

											 trailerCOM,//const btVector3& chassisCOM=btVector3(0.f,0.f,0.f)
											 trailerWheelOffset,//const btVector3& _chassisWheelOffset=btVector3(0.f,0.f,0.f),	// if != 0, you'd better set it using fraction of chassisLocalAABB components to position one of the most distant wheels of the vehicle
				 							 btVector3(0.f,0.f,0.f),//const btVector3& _wheelsAsymmetricOffset=btVector3(0.f,0.f,0.f)	// mainly useful to shift all the wheels together to the front/back

											 btScalar(0.5f),//const btScalar chassisFriction=btScalar(0.5f),
											 btScalar(0.0f),//const btScalar chassisRestitution=btScalar(0.0f),
                                             btScalar(2.0f),//const btScalar wheelFriction=btScalar(1.0f),
											 btScalar(0.0f),//const btScalar wheelRestitution=btScalar(0.0f),
											 // Tip: Set the following to lines to NULL when writing code, and reuse existing shapes when optimizing it (but it may create useless headaches...):
                                             NULL,//btCollisionShape* reuseThisCollisionShapeForChassis,	//If used, all other params must be set as if it was NULL
											 vehicle->getWheel(0) ? vehicle->getWheel(0)->getCollisionShape() : NULL,//btCollisionShape* reuseThisCollisionShapeForWheel,		//If used, all other params must be set as if it was NULL

											 -1,//const short int chassisCollisionFilterGroup=-1,
											 -1,//const short int chassisCollisionFilterMask=-1,
											 -1,//const short int wheelsCollisionFilterGroup=-1,
											 -1,//const short int wheelsCollisionFilterMask=-1,

											 // These should be left as default unless further tweaking is required...
											 btScalar(-1),//const btScalar verticalSlidersExtension=btScalar(-1),
											 btScalar(-1),//const btScalar verticalSlidersMotorVelocity=btScalar(-1),
											 btScalar(-1)//const btScalar verticalSlidersMotorMaxForce=btScalar(-1)
											);
				if (trailer)	{
					trailer->setVelocityLock(true);	//i.e. if we try to alter its velocity nothing happens.
					trailer->setSteeringLock(true);	//i.e. if we try to steer it nothing happens (probably useless, since we have used 0 steering wheels for it...).
				}

            }
            else printf("Error. I can't load \"truckTrailer.obj\"\n");
        }

	if (vehicle && trailer)	{
		// create contraint to link the two vehicles:
		const bool allowCollisions = false;	// "true" does not work as expected...
		const btScalar deltaHeight = allowCollisions ? chassisLocalAABB.z()*0.1f : 0;	// if we want to enable collisions between rigid bodies, to prevent the trailer from rotating past the part of the truck engine we'll add later on, we add an empty space between the 2 bodies...
                btVector3 pivotA(btVector3(-chassisCOM.x(),(-chassisCOM.y()-chassisLocalAABB.y()*0.3f)+deltaHeight*0.5f,-chassisCOM.z()-chassisLocalAABB.z()*0.5f+intersectionLength));
                btVector3 pivotB(-trailerCOM+btVector3(0,-trailerLocalAABB.y()*0.3625f-deltaHeight*0.5f,trailerLocalAABB.z()*0.5f-intersectionLength));

		btHingeConstraint* constraint = new btHingeConstraint(*vehicle->getChassis(), *trailer->getChassis(), pivotA,pivotB, chassisT.getBasis().getColumn(1),trailerT.getBasis().getColumn(1));//, bool useReferenceFrameA = false);
		m_dynamicsWorld->addConstraint(constraint, !allowCollisions);	// false = allow collisions between parts
		if (!allowCollisions) constraint->setLimit(-SIMD_HALF_PI,SIMD_HALF_PI,0.9f,0.3f,1.0f);//btScalar _softness = 0.9f, btScalar _biasFactor = 0.3f, btScalar _relaxationFactor = 1.0f);
	}

	}
	#endif //TEST_TRUCK


    #define TEST_FORKLIFT
    #ifdef TEST_FORKLIFT
    // Build Forklift
    {

    const btScalar chassisHeight(3);
    const btScalar wheelHeight(chassisHeight*0.5f);
    const btTransform chassisTransform(btQuaternion::getIdentity(),btVector3(-20.f,chassisHeight,0.f));
    const btScalar chassisMass(400);
    const btScalar wheelMass(20);
    const btScalar forksMass(30);

    btVector3 chassisFullAABBExtents,chassisCOMOffset,wheelFullAABBExtents,chassisWheelOffset;
    btTriangleMeshEx mesh;
    btHinge2Vehicle* vehicle = NULL;
    if (mesh.loadFromOBJFile("./Media/forklift/forklift.obj"))  {
        //mesh.center();	//Not here
        mesh.setHeight(chassisHeight);
        vehicleWheelMesh.setHeight(wheelHeight);
        chassisFullAABBExtents = mesh.getAabbFullExtents();
        wheelFullAABBExtents = vehicleWheelMesh.getAabbFullExtents();
        chassisCOMOffset.setValue(0.f,-(chassisHeight*0.5f)*0.8f,-(chassisFullAABBExtents.z()*0.5f)*0.25f);
        chassisWheelOffset.setValue((chassisFullAABBExtents.x()*0.5f)-wheelFullAABBExtents.x()*0.35f,-(chassisFullAABBExtents.y()*0.5f)+wheelFullAABBExtents.y()*0.55f,(chassisFullAABBExtents.z()*0.425f)-wheelFullAABBExtents.z()*0.5f);
        params.nClusters = 11;
        params.addFacesPoints = true;

        // Big stuff, I know, but note that almost all the params are in their default state...
        vehicle = localCreateHinge2Vehicle(
                                             chassisTransform,//const btTransform& startTransform,
                                             4,//const unsigned short numberOfWheels = 4,
                                             2,//const unsigned short numberOfSteeringWheels = 2,
                                             0,//const unsigned short tractionMode = 0,			// 0 = steering wheels, 1 = all the other wheels, 2 = all wheels ( NO DIFFERENTIAL IN ALL CASES )
                                             0,//const unsigned short wheelShapeTypeIndex = 2,	// 0 = cylinder, 1 = scaled sphere, 2 = very complex wheel shape, 3 = train wheel shape

                                             3,//const unsigned short wheelPositionMode = 0,	// 0 = NORMAL, 1 = PUT ALL WHEELS AT THE FRONT, 2 = PUT ALL WHEELS AT THE BACK, 3 = PLACE STEERING WHEELS AT THE BACK

                                             mesh,//const btTriangleMeshEx& chassisCenteredMesh,
                                             vehicleWheelMesh,//const btTriangleMeshEx& wheelCenteredMesh,

                                             params,//const btHACDCompoundShape::Params& chassisHACDCompoundShapeParams=btHACDCompoundShape::Params(),
                                             true,//const bool useFlatNormalsForChassisMesh= false,

                                             chassisMass,//const btScalar _chassisMass=btScalar(-1.f),
                                             wheelMass,//const btScalar _wheelMass=btScalar(-1),

                                             chassisCOMOffset,//const btVector3& chassisCOM=btVector3(0.f,0.f,0.f)
                                             chassisWheelOffset,//btVector3(chassisLocalAABB.x()*0.55,-1.5*chassisLocalAABB.y(),chassisLocalAABB.z()*0.4),//btVector3(0.f,0.f,0.f),//const btVector3& _chassisWheelOffset=btVector3(0.f,0.f,0.f),	// if != 0, you'd better set it using fraction of chassisLocalAABB components to position one of the most distant wheels (+X,-Y,+Z) of the vehicle
                                             btVector3(0.f,0.f,0.f),//const btVector3& _wheelsAsymmetricOffset=btVector3(0.f,0.f,0.f)	// mainly useful to shift all the wheels together to the front/back

                                             btScalar(0.5f),//const btScalar chassisFriction=btScalar(0.5f),
                                             btScalar(0.0f),//const btScalar chassisRestitution=btScalar(0.0f),
                                             btScalar(1.0f),//const btScalar wheelFriction=btScalar(1.0f),
                                             btScalar(0.0f),//const btScalar wheelRestitution=btScalar(0.0f),

                                             NULL,//btCollisionShape* reuseThisCollisionShapeForChassis,	//If used, all other params must be set as if it was NULL.  Use it only with chassis shapes previously created by this same method.
                                             NULL,//btCollisionShape* reuseThisCollisionShapeForWheel,		//If used, all other params must be set as if it was NULL.  Use it only with wheel shapes previously created by this same method.

                                             -1,//const short int chassisCollisionFilterGroup=-1,
                                             -1,//const short int chassisCollisionFilterMask=-1,
                                             -1,//const short int wheelsCollisionFilterGroup=-1,
                                             -1,//const short int wheelsCollisionFilterMask=-1,

                                             // These should be left as default unless further tweaking is required...
                                             wheelHeight*0.05f,//btScalar(-1),//const btScalar verticalSlidersExtension=btScalar(-1),
                                             btScalar(-1),//const btScalar verticalSlidersMotorVelocity=btScalar(-1),
                                             btScalar(-1)//const btScalar verticalSlidersMotorMaxForce=btScalar(-1)
                                             );

        if (vehicle) {
            m_driving_vehicles.push_back(vehicle); // enables driving input and camera targeting
            vehicle->setMaxPossibleVelocity(10.0f);		//35.0 is the default
            vehicle->setMinPossibleVelocity(-7.5f);
            vehicle->onBrake(); // stop it (otherwise inertia moves it a little)
        }
    }
    else printf("Error. Can't load file \"forklift.obj\"\n");

    bool loadOK=true;
    mesh.reset();
    if (vehicle && (loadOK = mesh.loadFromOBJFile("./Media/forklift/forks.obj")))	{
        mesh.setWidth(chassisFullAABBExtents.x()*2.0f/3.0f);
        const btVector3 centerPoint = mesh.getCenterPoint();
        mesh.center();
        const btVector3 meshHalfExtents = mesh.getAabbHalfExtents();
        mesh.shift(btVector3(0,+meshHalfExtents.y(),+meshHalfExtents.z()));	// Now 0 is in the bottom Z and bottom Y point (the way I want the COM to be)
        params.nClusters = 4;
        params.addFacesPoints = true;
        btCollisionShape* shape = new btHACDCompoundShape(mesh,params);
        if (shape)	{
            AssignMesh(shape,mesh,true);
            btTransform forksTransform(chassisTransform);
            forksTransform.setOrigin(forksTransform.getOrigin()+forksTransform.getBasis().getColumn(2)*(chassisFullAABBExtents*0.6575f));
            btRigidBody* forksBody = localCreateRigidBody(forksMass,forksTransform,shape);
            if (forksBody)	{
                forksBody->setFriction(0.75f);	// Not realistic...
                forksBody->setRestitution(0);

                // Add slider constraint
                //const btTransform localA(btQuaternion::getIdentity(),btVector3(0,-chassisCOMOffset.y(),chassisFullAABBExtents.z()*0.6575f));
                //const btTransform localB(btQuaternion::getIdentity(),btVector3(0,meshHalfExtents.y(),0));

                const btQuaternion q(btVector3(0,0,1),SIMD_HALF_PI);	// The slider seems to work around the X axis!
                const btTransform localA(q,btVector3(0,-chassisCOMOffset.y(),chassisFullAABBExtents.z()*0.6575f));	//Mmmhh, not sure why I don't have to rotate the origin around q...
                const btTransform localB(q,btVector3(0,meshHalfExtents.y(),0));										// Here too....

                btSliderConstraint*	m_forkSlider = new btSliderConstraint(*(vehicle->getChassis()), *forksBody, localA, localB, true);
                const btScalar lowerLimit = -wheelHeight*btScalar(3)/btScalar(10);	//-0.4f
                const btScalar upperLimit = chassisHeight*btScalar(5)/btScalar(6);  //2.5f
                const btScalar startingLockPoint = lowerLimit;	//We could just set it to 0.0f (or better half-way)
                m_forkSlider->setLowerLinLimit(startingLockPoint);
                m_forkSlider->setUpperLinLimit(startingLockPoint);
        //		m_forkSlider->setLowerAngLimit(-LIFT_EPS);
        //		m_forkSlider->setUpperAngLimit(LIFT_EPS);
                m_forkSlider->setLowerAngLimit(0.0f);
                m_forkSlider->setUpperAngLimit(0.0f);
                m_dynamicsWorld->addConstraint(m_forkSlider, false);//true);

                // Now we add a helper truct to manipulate it better:
                btSliderHelperStruct* shs = new btSliderHelperStruct(m_forkSlider,
                                                                        lowerLimit,
                                                                        upperLimit,
                                                                        8.0f,
                                                                        1.75f
                                                                     );
                g_sliderHelperStructs.push_back(shs);
                SetUserPointer(vehicle->getChassis(),(void*)(g_sliderHelperStructs.size()));	// Dirty stuff (We store (index + 1) !!)
            }
        }

        // Optional: Build some blocks to play with:
        {
            btCompoundShape* shape = NULL;
            btTransform T;T.setIdentity();
            const btScalar hx(chassisFullAABBExtents.x()*0.875f);
            const btScalar hy(hx*0.5f);
            const btScalar hz(hx*btScalar(1)/btScalar(2));

            btTriangleMeshEx mesh;
            if (mesh.loadFromFile("./Media/wooddenBlock/wooddenBlock.obj"))	{
                mesh.center();
                mesh.setHalfExtents(btVector3(hx,hy,hz));
                params.nClusters = 3;
                params.addFacesPoints = false;
                shape = new btHACDCompoundShape(mesh,params);
                this->AssignMesh(shape,mesh,true);
            }
            else	{
                shape = new btCompoundShape();
                shape->addChildShape(T,new btBoxShape(btVector3(hx,hy,hz)));
                btBoxShape* shape2 = new btBoxShape(btVector3(hx/btScalar(14),hy*1.4f,hz*1.4f));
                T.setOrigin(btVector3(-hx*btScalar(8)/btScalar(10),0,0));
                shape->addChildShape(T,shape2);
                T.setOrigin(btVector3(hx*btScalar(8)/btScalar(10),0,0));
                shape->addChildShape(T,shape2);
            }

            const btScalar mass(80);
            const btScalar friction(2);
            const btScalar restitution(0);

            const int numBlocksInARow(10);
            for (int i = 0; i < numBlocksInARow;i++)	{
                btVector3& v = T.getOrigin();
                v = chassisTransform.getOrigin();
                v.setZ(v.z()+15);
                v.setX(v.x()+hx*3*(btScalar)(i-numBlocksInARow/2));

                btRigidBody* body = localCreateRigidBody(mass,T,shape);
                body->setFriction(friction);
                body->setRestitution(restitution);


                v.setY(v.y()+hy*3);
                body = localCreateRigidBody(mass,T,shape);
                body->setFriction(friction);
                body->setRestitution(restitution);
            }

        }
    }
    else if (!loadOK) printf("Error. Can't load file \"forks.obj\"\n");
    }
    #endif // TEST_FORKLIFT


    #define TEST_TRAIN
    #define NUM_TRAINS 2    // >0 !
    const static size_t numWagonsPerTrain[]={12,8,4,0,6};   // in [0,35] (TODO: test performance with many wagons)
     // Things to do to make the trains stay into the tracks:
    /*
    1- compile Bullet And This Demo with BT_USE_DOUBLE_PRECISION (a must for Linux!)
    2- #define BETTER_STABILITY_OF_WHEEL_TRAIN_SHAPES 1 above
    3- play with LooseWheelsHalfAngles
    4- change the COM offset of the locomotive (and optionally of the wagons)
    5- reduce the maxAngularVelocity of the locomotive (of course)
    6- make the track curves less sharp, and make the tracks higher (this should make the wheel train shapes higher too, and you can/must reduce maxAngularVelocity to get the same linear speed)
    7- dynamically adjust the applied speed in different parts of the tracks (hehe... I leave this for others!)
    That's all, and it's still not easy to make the train run very fast in the tracks!
    P.S. what about increasing the Bullet time step or overriding the number of the constraint iterations of the Hinge2Constraint ?
    NOTE: tracks near the center have steeper curves.
    NOTE: using a lower number of wagons does not seem to help stability a lot (at least in my case). I suggest using no wagons while experimenting (less params to tweak and then it should work with wagons too).
    NOTE: of course if you add 100 wagons, there will be no room to position them on the track...
    */
    #ifdef TEST_TRAIN
	// Build Train
	{
        btTriangleMeshEx trainWheelMesh;
        if (trainWheelMesh.loadFromOBJFile("./Media/train/wheel/trainWheel.obj"))  {
    		trainWheelMesh.center();
    		//trainWheelMesh.rotate(btQuaternion(btVector3(0,1,0),SIMD_PI));
		}
        else printf("Error. Can't load file \"trainWheel.obj\"\n");

		btTriangleMeshEx trainLocomotiveMesh;
        if (trainLocomotiveMesh.loadFromOBJFile("./Media/train/locomotive/trainLocomotive.obj"))  {
    		trainLocomotiveMesh.center();
    		trainLocomotiveMesh.rotate(btQuaternion(btVector3(0,1,0),SIMD_PI));
		}
        else printf("Error. Can't load file \"trainLocomotive.obj\"\n");

		btTriangleMeshEx trainWagon1Mesh;
        if (trainWagon1Mesh.loadFromOBJFile("./Media/train/wagon/trainWagon.obj"))  {
    		trainWagon1Mesh.center();
    		trainWagon1Mesh.rotate(btQuaternion(btVector3(0,1,0),SIMD_PI));
		}
        else printf("Error. Can't load file \"trainWagon.obj\"\n");

		btTriangleMeshEx trainWagon2Mesh;
        if (trainWagon2Mesh.loadFromOBJFile("./Media/train/wagon2/trainWagon2.obj"))  {
    		trainWagon2Mesh.center();
    		trainWagon2Mesh.rotate(btQuaternion(btVector3(0,1,0),SIMD_PI));
		}
        else printf("Error. Can't load file \"trainWagon2.obj\"\n");

        const btScalar trainLocomotiveMass(400);
        const btScalar trainLocomotiveWheelMass(15);
        const btScalar trainWagonWheelMass(10);
        const btScalar trainWagonMass(100);

        //---------------------------------------------------------------------------------------------------------------------------------------------------------------
        //---------------------------------------------------------------------------------------------------------------------------------------------------------------
		// Build railway
		const btScalar groundLevel = 0.0f;
        const btScalar railwayHeight = 0.3f;
        const btScalar railwayWidth = 0.15f;	// btWheelTrainShape innerWidth must equal this value
		const btScalar railwayDistance = 3.0f; 	// trainWidth must equal this value
        const btScalar railwayFriction = 0.25f;//0.5f;	// Too high can make the train get out of wheels
        btScalar railwayRadius = 400.0f * 0.98f;    // Very important: it determines the overall size of the circuit, the max number of wagons, ( and maybe the possibility of using Bullet in single precision under Windows, since Linux always needs double precison to make the train stay in the tracks )
                                                    // With 12 wagon max, we can use   200.0f * 0.98f;
                                                    // with 35 wagons max, we can use  400.0f * 0.98f;
        //-------------warnings-------------------------------------------------------------------------------------------------
        #ifdef BT_USE_DOUBLE_PRECISION
        #ifdef _WIN32
        printf("Warning: the train might not work or go out of the tracks easily when using Windows with BT_USE_DOUBLE_PRECISION (at least in my tests so far).\n");
        #endif //_WIN32
        #else
        #ifndef _WIN32
        printf("Warning: the train might go out of the tracks easily when not using BT_USE_DOUBLE_PRECISION (for both Bullet and this demo) in a non-Windows OS (at least in my tests so far).\n");
        #endif //_WIN32
        #endif
        //-----------------------------------------------------------------------------------------------------------------------

		const btScalar D(railwayRadius);
        const btScalar L(railwayRadius*0.75);   // Actually THIS determines the max number of wagons, but it's linked to railwayRadius ATM...
		const btScalar railwayLength = L;	// I guess this is used just to add a larger number of wagons... we just set it to a high number for now...
		btAlignedObjectArray <btVector3> points;
        // Build railway by filling 'points':
        #define RAILWAY_BUILD 2 // in [1,2]
        {
        #if RAILWAY_BUILD==2    // Oval Railway with 4 straight sections
            const btScalar L2(L*0.5);
            railwayRadius+=L2;  //railwayRadius is used later to position the train on the tracks
            const btScalar HL2 = L2*0.5;
            const int numSegments(54);  // For a 360 degree curve
            for (int i=0;i<=numSegments/4;i++)                  points.push_back(btVector3(D*btCos(i*2.0f*SIMD_PI/numSegments),0,D*btSin(i*2.0f*SIMD_PI/numSegments))+btVector3(L2,0,L));
            for (int i=numSegments/4;i<=numSegments/2;i++)      points.push_back(btVector3(D*btCos(i*2.0f*SIMD_PI/numSegments),0,D*btSin(i*2.0f*SIMD_PI/numSegments))+btVector3(-L2,0,L));
            for (int i=numSegments/2;i<=3*numSegments/4;i++)    points.push_back(btVector3(D*btCos(i*2.0f*SIMD_PI/numSegments),0,D*btSin(i*2.0f*SIMD_PI/numSegments))+btVector3(-L2,0,-L));
            for (int i=3*numSegments/4;i<=numSegments;i++)      points.push_back(btVector3(D*btCos(i*2.0f*SIMD_PI/numSegments),0,D*btSin(i*2.0f*SIMD_PI/numSegments))+btVector3(L2,0,-L));
        #else // Default Oval Railway ( 2 straight sections only )
            const int numSegments(54);  // For a 360 degree curve
            // Oval circuit. Please note that curves can't be too sharp (we use numSegments WIDE segments for a 360 degrees turn).
            for (int i=0;i<=numSegments/2;i++) points.push_back(btVector3(D*btCos(i*2.0f*SIMD_PI/numSegments),0,D*btSin(i*2.0f*SIMD_PI/numSegments))+btVector3(0,0,L));
            for (int i=numSegments/2;i<=numSegments;i++) points.push_back(btVector3(D*btCos(i*2.0f*SIMD_PI/numSegments),0,D*btSin(i*2.0f*SIMD_PI/numSegments))+btVector3(0,0,-L));
        #endif //RAILWAY_BUILD
        }
        const int numPoints = points.size();
		//----- Optional: build more than one parallel track------
                const int numOfRailwaySets(NUM_TRAINS);		// MUST BE >=0! 		We can build more than one parallel track
		const btScalar distanceBetweenTwoRailSets(railwayDistance*0.75f);
        int platformNumber = NUM_TRAINS;	// For positioning the train. Twickable in [1,numOfRailwaySets]
		//
		if (platformNumber<1) platformNumber=1;
		else if (platformNumber>numOfRailwaySets) platformNumber=numOfRailwaySets;
		// We must set 'railwayRadius' just because 'railwayRadius' is used later to position the train on the railway correctly:
		const bool oddNumberOfSets(numOfRailwaySets%2);
		if (oddNumberOfSets) {
			platformNumber = platformNumber - numOfRailwaySets/2 -1;
			railwayRadius = railwayRadius + (distanceBetweenTwoRailSets + railwayDistance)*platformNumber;
		}
		else	{
			platformNumber = (platformNumber - numOfRailwaySets/2);
			railwayRadius = railwayRadius - (distanceBetweenTwoRailSets*0.5f + railwayDistance*0.5f) + (distanceBetweenTwoRailSets + railwayDistance)*platformNumber;
		}
		//---------------------------------------------------------
        btRailwayCompoundShape* shape = new btRailwayCompoundShape(railwayDistance,railwayWidth,railwayHeight,numOfRailwaySets,distanceBetweenTwoRailSets);
		shape->addPoints(&points[0],numPoints);
        shape->connectEndToStart();
        //shape->displayDebugInfo();
        shape->AssignDefaultBoxMeshForRailway(this,true);   // Optional: Assign default mesh to the railway box shaped pieces

        const btQuaternion railwayOrientation(btVector3(0,1,0),btRadians(-40.75));
        const btTransform T(railwayOrientation,btVector3(0,groundLevel,0));
        localCreateCollisionObjectByUnwrappingCompoundShape(T,shape,railwayFriction,0); // is this faster than using a single object with btRailwayCompoundShape? Don't know!
        delete shape;shape=NULL;    // When the method above is called, the compound is removed and its child shapes become N new collision objects.
        //---------------------------------------------------------------------------------------------------------------------------------------------------------------
        //---------------------------------------------------------------------------------------------------------------------------------------------------------------

        // Stuff used later (mostly optional, can be ignored for now)
        const size_t numWagonsPerTrainSize = sizeof(numWagonsPerTrain)/sizeof(numWagonsPerTrain[0]);
        static Material trainMaterials[5];   // override materials must be persistent (because AFAIR they're not copied)
        trainMaterials[0]= Material(0.2f,0.5f,0.2f,false,0.4f,0.4f,4.0f/128.0f);
        trainMaterials[1]= Material(0.2f,0.2f,0.5f,false,0.4f,0.4f,4.0f/128.0f);
        trainMaterials[2]= Material(0.4f,0.1f,0.1f,false,0.4f,0.4f,4.0f/128.0f);
        trainMaterials[3]= Material(0.5f,0.5f,0.2f,true,0.4f,0.1f,4.0f/128.0f);
        trainMaterials[4]= Material(0.4f,0.4f,0.8f,false,0.4f,0.4f,4.0f/128.0f);
        const size_t numOvrMaterials = sizeof(trainMaterials)/sizeof(trainMaterials[0]);  //can be any number (i.e. we can add materials here)

        //---------------------------------------------------------------------------------------------------------------------------------------------------------------
        //---------------------------------------------------------------------------------------------------------------------------------------------------------------
        // Build vehicles
		btAlignedObjectArray< btHinge2Vehicle* > trainVehicles;

		trainLocomotiveMesh.setWidth(railwayDistance);

			//Wheel
            const btScalar	wheelOuterRadius = railwayHeight        *1.4f;//1.5f//*0.f;
            const btScalar	wheelTotalWidth = railwayWidth*3.5f;

        const btVector3 wheelLocalAABB(wheelTotalWidth,wheelOuterRadius*2,wheelOuterRadius*2);
		trainWheelMesh.setFullExtents(wheelLocalAABB);


    // locomotive data
	const btVector3 locomotiveLocalAABB(trainLocomotiveMesh.getAabbFullExtents());
    const btVector3 locomotiveCOMOffset(0.f,-locomotiveLocalAABB.y()*0.4f,0.f);//-locomotiveLocalAABB.z()*0.15f);
	const btVector3 locomotiveWheelOffset(railwayDistance*0.5f,-locomotiveLocalAABB.y()*0.5f-wheelOuterRadius*0.5f,locomotiveLocalAABB.z()*0.45f-2*wheelOuterRadius);
    const btScalar locomotiveLooseTractionWheelsHalfAngle = 1.5f;  // >=0. 	// Basically allows passive steering of wheels (maybe this helps in sharp curves... who knows. I don't think they are necessary here, though)
    const btScalar locomotiveLooseFrontWheelsHalfAngle = 5.0f;     // >=0.


    btCollisionShape* locomotiveChassisShape = NULL; // to reuse this if multiple locomotives are needed.
    btCollisionShape* trainWheelShape = NULL;        // to reuse this in future calls (for wagons too)

    // wagons data
    trainWagon1Mesh.setWidth(railwayDistance);//*1.1f);
    trainWagon2Mesh.setWidth(railwayDistance);//*1.1f);
    const btVector3 wagon1LocalAABB(trainWagon1Mesh.getAabbFullExtents());
    const btVector3 wagon2LocalAABB(trainWagon2Mesh.getAabbFullExtents());
    const btVector3 wagon1COMOffset(0.f,-wagon1LocalAABB.y()*0.35f,0.f);
    const btVector3 wagon2COMOffset(0.f,-wagon2LocalAABB.y()*0.35f,0.f);
    const btVector3 wagon1WheelOffset(railwayDistance*0.5f,-wagon1LocalAABB.y()*0.5f-wheelOuterRadius*0.5f,wagon1LocalAABB.z()*0.5f-wheelOuterRadius);
    const btVector3 wagon2WheelOffset(railwayDistance*0.5f,-wagon2LocalAABB.y()*0.5f-wheelOuterRadius*0.5f,wagon2LocalAABB.z()*0.5f-wheelOuterRadius);
    const btScalar wagonLooseWheelsHalfAngle = 2.5f;   // >=0.

    btCollisionShape* reuseWagon1Shape(NULL);
    btCollisionShape* reuseWagon2Shape(NULL);
    const btScalar baseRailwayRadius = railwayRadius;

    // Main loop:
    for (size_t trainIndex=0;trainIndex<NUM_TRAINS;trainIndex++)
    {
    //if (trainIndex==2) return;
    //printf("trainIndex=%d\n",trainIndex);
    const bool reverseTrack = (trainIndex/2)%2==1;
    const btScalar reverseTrackFactor = reverseTrack ? -1 : 1;

    const bool reverseOrientation =  !reverseTrack ? trainIndex%2==1 : trainIndex%2==0;
    const btScalar reverseOrientationFactor = reverseOrientation ? -1 : 1;
    //-------------------------------------------------------------------------------------------------
        railwayRadius = baseRailwayRadius - trainIndex*(distanceBetweenTwoRailSets + railwayDistance);// For positioning the train. railwayRadius was set for platFormNumber = NUM_TRAINS, now we use platFormNumber = NUM_TRAINS-trainIndex, so we subtract the required quantity
    //---------------------------------------------------------------------------------------------------
    btVector3 locomotiveOriginWithoutRailwayOrientation = btVector3(reverseTrackFactor*railwayRadius,locomotiveLocalAABB.y()*0.5f+groundLevel+railwayHeight+5.0f*wheelOuterRadius,reverseOrientationFactor*(railwayLength*0.5f-1.0f*locomotiveLocalAABB.z()));//-chassisLength*0.55f);
    btQuaternion locomotiveOrientationWithoutRailwayOrientation = !reverseOrientation ? btQuaternion::getIdentity() : btQuaternion(btVector3(0,1,0),btRadians(180));

    btTransform locomotiveT(railwayOrientation*locomotiveOrientationWithoutRailwayOrientation,locomotiveOriginWithoutRailwayOrientation);
    locomotiveT.getOrigin() = quatRotate(railwayOrientation , locomotiveT.getOrigin());


    params.nClusters = 5;
	params.addFacesPoints = true;
	//params.displayDebugInfo=true;
	btHinge2Vehicle* vehicle = localCreateHinge2Vehicle(
											 locomotiveT,//const btTransform& startTransform,
											 4,//const unsigned short numberOfWheels = 4,
											 2,//const unsigned short numberOfSteeringWheels = 2,
											 1,//const unsigned short tractionMode = 0,			// 0 = steering wheels, 1 = not steering wheels, 2 = all wheels ( NO DIFFERENTIAL IN ALL CASES )
											 3,//const unsigned short wheelShapeTypeIndex = 2,	// 0 = cyl, 1 = scaled sphere, 2 = very complex wheel shape, 3 = train wheel shape

											 0,//const unsigned short wheelPositionMode = 0,	// 0 = NORMAL, 1 = PUT ALL WHEELS AT THE FRONT, 2 = PUT ALL WHEELS AT THE BACK

											trainLocomotiveMesh,
											trainWheelMesh,

											 params,//const btHACDCompoundShape::Params& chassisHACDCompoundShapeParams=btHACDCompoundShape::Params(),
											 true,//const bool useFlatNormalsForChassisMesh= false,

                                                                                         trainLocomotiveMass,//const btScalar _chassisMass=btScalar(-1.f),
                                                                                         trainLocomotiveWheelMass,//const btScalar _wheelMass=btScalar(-1),

											 locomotiveCOMOffset,//const btVector3& chassisCOM=btVector3(0.f,0.f,0.f)
											 locomotiveWheelOffset,//const btVector3& _chassisWheelOffset=btVector3(0.f,0.f,0.f),	// if != 0, you'd better set it using fraction of chassisLocalAABB components to position one of the most distant wheels of the vehicle
				 							 btVector3(0.f,0.f,0.f),//const btVector3& _wheelsAsymmetricOffset=btVector3(0.f,0.f,0.f)	// mainly useful to shift all the wheels together to the front/back

											 btScalar(0.5f),//const btScalar chassisFriction=btScalar(0.5f),
											 btScalar(0.0f),//const btScalar chassisRestitution=btScalar(0.0f),
                                             btScalar(0.2f),//const btScalar wheelFriction=btScalar(1.0f),
											 btScalar(0.0f),//const btScalar wheelRestitution=btScalar(0.0f),
											 // Tip: Set the following to lines to NULL when writing code, and reuse existing shapes when optimizing it (but it may create useless headaches...):
                                             locomotiveChassisShape,//btCollisionShape* reuseThisCollisionShapeForChassis,	//If used, all other params must be set as if it was NULL
                                             trainWheelShape,//btCollisionShape* reuseThisCollisionShapeForWheel,		//If used, all other params must be set as if it was NULL

											 -1,//const short int chassisCollisionFilterGroup=-1,
											 -1,//const short int chassisCollisionFilterMask=-1,
											 -1,//const short int wheelsCollisionFilterGroup=-1,
											 -1,//const short int wheelsCollisionFilterMask=-1,

											 // These should be left as default unless further tweaking is required...
                                             btScalar(-1),//const btScalar verticalSlidersExtension=btScalar(-1),
                                             btScalar(-1),//const btScalar verticalSlidersMotorVelocity=btScalar(-1),
                                             btScalar(-1)//const btScalar verticalSlidersMotorMaxForce=btScalar(-1)
											);

	if (vehicle) {
        if (!locomotiveChassisShape) locomotiveChassisShape = vehicle->getChassis()->getCollisionShape();
        if (!trainWheelShape && vehicle->getNumWheels()>0) trainWheelShape = vehicle->getWheel(0)->getCollisionShape();
        //-----------------------------------------------------------------------------------------------------
		// Problem: the locomotive engine wheels skid too much. To make the locomotive go faster,
		// we try to increase the friction of the "motorized" wheels only, leaving the others as they are.
		// (I guess that decreasing the wagon mass would have worked too...)
		const int numTractionWheels = vehicle->getNumTractionWheels();
		for (int i = 0; i < numTractionWheels; i++)	{
			btRigidBody* wheel = vehicle->getTractionWheel(i);
            if (wheel) wheel->setFriction(20.0f);
        }
        //------------------------------------------------------------------------------------------------------
        const btScalar maxAngularVelocity = 70.0f;	// It takes a while to get to full speed (and the train can go out from the rails if it's too high)
		vehicle->setMaxPossibleVelocity(maxAngularVelocity);		//35.0f is the class default ATM
        vehicle->setMinPossibleVelocity(-maxAngularVelocity);	//-MaxPossibleVelocity*0.3f is the class default ATM
        //vehicle->setEngineMotorForce(20.0);
        //------------------------------------------------------------------------------------------------------
        if (locomotiveLooseFrontWheelsHalfAngle>0)      vehicle->setLooseWheels(btRadians(-locomotiveLooseFrontWheelsHalfAngle),btRadians(+locomotiveLooseFrontWheelsHalfAngle),0); // 0 = front wheels, 1 = rear wheels, 2 = all wheels (default)
        if (locomotiveLooseTractionWheelsHalfAngle>0)   vehicle->setLooseWheels(btRadians(-locomotiveLooseTractionWheelsHalfAngle),btRadians(+locomotiveLooseTractionWheelsHalfAngle),1); // 0 = front wheels, 1 = rear wheels, 2 = all wheels (default)
        //-------------------------------------------------------------------------------------------------------
        vehicle->setVelocity( // This is the angular velocity of the spinning wheels, whose radius must be equal
						maxAngularVelocity,//btScalar targetVelocity = 5.0f,	// final angular velocity of the spinning wheels
                        15.0f//btScalar maxMotorForce = 0.1f		// strength of the motor to get to it (i.e. gear or something like that)
                        );
		vehicle->setSteeringLock(true);	// So the user can't make the wheels steer

        trainVehicles.clear();
		trainVehicles.push_back(vehicle);	// Used later for creating the constraints that will link the vehicles
		m_driving_vehicles.push_back(vehicle); // enables driving input and camera targeting

        // Optional: color customization
        {
            btRigidBody* trainChassis = vehicle->getChassis();
            if (trainChassis) {
                btTriangleMeshExFootPrint::Overrides& ovr = GetMeshOverrides(*trainChassis);
                ovr.materialArray.push_back(NULL);                                           // NULL can be used to skip overriding a material
                ovr.materialArray.push_back(NULL);
                ovr.materialArray.push_back(&trainMaterials[trainIndex%numOvrMaterials]);    // Here we override mesh material[2] only (if it exists)
                // Please note that ovr.materialArray.size() can be any number (unrelated to the assigned mesh number of materials) and it won't crash the application (only valid materials will be overridden).
            }
        }

	}
    else continue;

    const size_t numberOfWagons = trainIndex<numWagonsPerTrainSize?numWagonsPerTrain[trainIndex]:numWagonsPerTrain[numWagonsPerTrainSize-1];// Well, it should depend on the max straight railway segment length...
	for (int t = 0; t < numberOfWagons;t++)	{
        //printf("wagonIndex=%d\n",t);
        //---------------------------------------------------------------------
        const bool isOdd(t%2==1);	//TODO: randomize it and seeif it works
        const btTriangleMeshEx& wagonMesh = isOdd ? trainWagon1Mesh : trainWagon2Mesh;
        const btVector3& wagonLocalAABB = isOdd ? wagon1LocalAABB : wagon2LocalAABB;
        const btVector3& wagonCOMOffset = isOdd ? wagon1COMOffset : wagon2COMOffset;
        const btVector3& wagonWheelOffset = isOdd ? wagon1WheelOffset : wagon2WheelOffset;
        btCollisionShape*& reuseWagonShape = isOdd ? reuseWagon1Shape : reuseWagon2Shape;

        if (isOdd)	{
			params.nClusters = 2;
			params.addFacesPoints = false;
			//params.displayDebugInfo=true;
		}
		else	{
			params.nClusters = 5;
			params.addFacesPoints = true;
			//params.displayDebugInfo=true;
		}

        btVector3 wagonTWithoutRailwayOrientation(reverseTrackFactor*railwayRadius,wagonLocalAABB.y()*0.5f+groundLevel+railwayHeight+5.0f*wheelOuterRadius,locomotiveOriginWithoutRailwayOrientation.z()
                                                  -reverseOrientationFactor*(locomotiveLocalAABB.z()+wagonLocalAABB.z()*1.2f*t));

        btTransform T(railwayOrientation*locomotiveOrientationWithoutRailwayOrientation,wagonTWithoutRailwayOrientation);
        T.getOrigin() = quatRotate(railwayOrientation , T.getOrigin());


        vehicle = localCreateHinge2Vehicle(
											 T,//const btTransform& startTransform,
											 4,//const unsigned short numberOfWheels = 4,
											 2,//const unsigned short numberOfSteeringWheels = 2,
											 0,//const unsigned short tractionMode = 0,			// 0 = steering wheels, 1 = not steering wheels, 2 = all wheels ( NO DIFFERENTIAL IN ALL CASES )
											 3,//const unsigned short wheelShapeTypeIndex = 2,	// 0 = cyl, 1 = scaled sphere, 2 = very complex wheel shape, 3 = train wheel shape

											 0,//const unsigned short wheelPositionMode = 0,	// 0 = NORMAL, 1 = PUT ALL WHEELS AT THE FRONT, 2 = PUT ALL WHEELS AT THE BACK

											 wagonMesh,
											 trainWheelMesh,

											 params,
											 true,

                                                                                         trainWagonMass,//const btScalar _chassisMass=btScalar(-1.f),
                                                                                         trainWagonWheelMass,//const btScalar _wheelMass=btScalar(-1),

											 wagonCOMOffset,//const btVector3& chassisCOM=btVector3(0.f,0.f,0.f)
											 wagonWheelOffset,//locomotiveWheelOffset,//const btVector3& _chassisWheelOffset=btVector3(0.f,0.f,0.f),	// if != 0, you'd better set it using fraction of chassisLocalAABB components to position one of the most distant wheels of the vehicle
				 							 btVector3(0.f,0.f,0.f),//const btVector3& _wheelsAsymmetricOffset=btVector3(0.f,0.f,0.f)	// mainly useful to shift all the wheels together to the front/back

											 btScalar(0.5f),//const btScalar chassisFriction=btScalar(0.5f),
											 btScalar(0.0f),//const btScalar chassisRestitution=btScalar(0.0f),
											 btScalar(0.2f),//const btScalar wheelFriction=btScalar(1.0f),
											 btScalar(0.0f),//const btScalar wheelRestitution=btScalar(0.0f),
											 // Tip: Set the following to lines to NULL when writing code, and reuse existing shapes when optimizing it (but it may create useless headaches...):
											 reuseWagonShape,//t > 0 ? vehicle->getChassis() ? vehicle->getChassis()->getCollisionShape() : NULL : NULL,//btCollisionShape* reuseThisCollisionShapeForChassis,	//If used, all other params must be set as if it was NULL
                                             trainWheelShape,//(vehicle && vehicle->getWheel(0)) ? vehicle->getWheel(0)->getCollisionShape() : NULL,//btCollisionShape* reuseThisCollisionShapeForWheel,		//If used, all other params must be set as if it was NULL

											 -1,//const short int chassisCollisionFilterGroup=-1,
											 -1,//const short int chassisCollisionFilterMask=-1,
											 -1,//const short int wheelsCollisionFilterGroup=-1,
											 -1,//const short int wheelsCollisionFilterMask=-1,

											 // These should be left as default unless further tweaking is required...
                                             btScalar(-1),//const btScalar verticalSlidersExtension=btScalar(-1),
											 btScalar(-1),//const btScalar verticalSlidersMotorVelocity=btScalar(-1),
											 btScalar(-1)//const btScalar verticalSlidersMotorMaxForce=btScalar(-1)
											);

	if (vehicle) {
        if (!reuseWagon1Shape && isOdd && vehicle->getChassis()) reuseWagon1Shape = vehicle->getChassis()->getCollisionShape();
        if (!reuseWagon2Shape && !isOdd && vehicle->getChassis()) reuseWagon2Shape = vehicle->getChassis()->getCollisionShape();

        if (wagonLooseWheelsHalfAngle>0) vehicle->setLooseWheels(btRadians(-wagonLooseWheelsHalfAngle),btRadians(+wagonLooseWheelsHalfAngle));

        vehicle->setSteeringLock(true);	// So the user can't make the wheels steer
		vehicle->setVelocityLock(true);	// So the user can't make the "engine" wheels turn

        trainVehicles.push_back(vehicle);   // used later for constraints between vehicles

        // Optional: color customization
        {
            btRigidBody* wagonChassis = vehicle->getChassis();
            if (wagonChassis) {
                btTriangleMeshExFootPrint::Overrides& ovr = GetMeshOverrides(*wagonChassis);
                ovr.materialArray.push_back(&trainMaterials[trainIndex%numOvrMaterials]);       // Here we override mesh material[0] only (if it exists), and leave the others as they are
                // Please note that ovr.materialArray.size() can be any number (unrelated to the assigned mesh number of materials) and it won't crash the application (only valid materials will be overridden).
            }
        }
	}
	else continue;

    //#define WAGONS_WITHOUT_CONSTRAINTS
    #ifndef WAGONS_WITHOUT_CONSTRAINTS
       //Wagon constraint Constraint-------------------------------------------------------------
		btRigidBody* bodyA = trainVehicles[t]->getChassis();
		btRigidBody* bodyB = trainVehicles[t+1]->getChassis();


		btTransform pivotA=btTransform::getIdentity();
		btTransform pivotB=btTransform::getIdentity();
		if (t==0) {
			pivotA.setOrigin(-locomotiveCOMOffset+btVector3(0,-locomotiveLocalAABB.y()*0.5,-wagonLocalAABB.z()*0.6));
			pivotB.setOrigin(-wagonCOMOffset+btVector3(0,-wagonLocalAABB.y()*0.5,wagonLocalAABB.z()*0.6));
		}
		else {
			pivotA.setOrigin(-wagonCOMOffset+btVector3(0,-wagonLocalAABB.y()*0.5,-wagonLocalAABB.z()*0.6));
			pivotB.setOrigin(-wagonCOMOffset+btVector3(0,-wagonLocalAABB.y()*0.5,wagonLocalAABB.z()*0.6));
		}

		btPoint2PointConstraint* constraint = new btPoint2PointConstraint(*bodyA, *bodyB, pivotA.getOrigin(), pivotB.getOrigin());
		//btConeTwistConstraint* constraint = new btConeTwistConstraint(*bodyA, *bodyB, pivotA, pivotB);
		//constraint->setLimit(M_PI_2, M_PI_2, 0);	//??? I have to learn more about it....

		m_dynamicsWorld->addConstraint(constraint, false);
		constraint->setDbgDrawSize(wagonLocalAABB.z()*0.1f);
        //--------------------------------------------------------------------------
    #endif //WAGONS_WITHOUT_CONSTRAINTS

    }  // wagon loop

    } // Main Train loop

    //endregion
	}
	#endif //TEST_TRAIN




}


void Hinge2VehicleDemo::preTickCallback(btScalar timeStep)	{

	IDrivingInterface* vehicle = m_activeVehicleIndex>=0 &&  m_activeVehicleIndex<m_driving_vehicles.size() ? m_driving_vehicles[m_activeVehicleIndex] : NULL;
	if (vehicle ) {
		if (g_vehicleKeyPressedLeft) vehicle->onLeft();
		if (g_vehicleKeyPressedRight) vehicle->onRight();
		if (g_vehicleKeyPressedUp) vehicle->onUp();
		if (g_vehicleKeyPressedDown) vehicle->onDown();
		if (g_vehicleKeyPressedSpecialActionUp) {
			if (vehicle->getChassis())	{
            	const int indexPlusOne =  (const long int) GetUserPointer(vehicle->getChassis());
				if (indexPlusOne>0 && indexPlusOne>=g_sliderHelperStructs.size())	{
					const btSliderHelperStruct* shs = g_sliderHelperStructs[indexPlusOne-1];
					if (shs) shs->slideUp();
				}
			}
		}
		if (g_vehicleKeyPressedSpecialActionDown) {
			if (vehicle->getChassis())	{
            	const int indexPlusOne =  (const long int) GetUserPointer(vehicle->getChassis());
				if (indexPlusOne>0 && indexPlusOne>=g_sliderHelperStructs.size())	{
					const btSliderHelperStruct* shs = g_sliderHelperStructs[indexPlusOne-1];
					if (shs) shs->slideDown();
				}
			}
		}

		vehicle->onPostInput();
	}
}

void Hinge2VehicleDemo::clientMoveAndDisplay() {

	 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	static btScalar bulletFixedTimeStep(btScalar(1.)/btScalar(60.));
	static unsigned long frame(0);	// A simple frame counter
	++frame;


	float dt = (getDeltaTimeMicroseconds() * 0.000001);
    if (dt>6*bulletFixedTimeStep && frame>5)	{
		printf("Error. CPU is not fast enough to produce physic support at frame n.%d\n",frame);
		dt=0;
	}
	else m_dynamicsWorld->stepSimulation(dt,5,bulletFixedTimeStep);



    m_dynamicsWorld->debugDrawWorld();

	renderme();

    glFlush();
    swapBuffers();

}

void Hinge2VehicleDemo::exitPhysics() {
	//cleanup in the reverse order of creation/initialization

	// cleanup vehicles (actually it cleans up the constraints only, so probably it could be skipped, but we may add additional stuff inside vehicles...)
	for (int i=0;i<m_hinge2Vehicles.size();i++) m_hinge2Vehicles[i]->destroy();
	m_hinge2Vehicles.clear();
    m_driving_vehicles.clear();m_activeVehicleIndex=0;	// Just in case someone wants to start up again... by pressing SPACE


	GlutDemoApplicationEx::exitPhysics();

	for (int i=0;i<g_sliderHelperStructs.size();i++)	{
		btSliderHelperStruct* m = g_sliderHelperStructs[i];
		if (m)	{
			delete m;m=NULL;
		}
	}
	g_sliderHelperStructs.resize(0);

}

 //region Keyboard Stuff
void Hinge2VehicleDemo::keyboardCallback(unsigned char key, int x, int y){
    if (key=='a' || key=='f') return;
	if (key=='r') {
		g_renderMode++;
		if (g_renderMode==3) g_renderMode=0;
        if (g_renderMode==0) g_renderMode++;	// Optional: skips btShapeHull view for all objects (Bullet default view) in favour of btShapeHull + btConvexHullComputer in case of btConvexHullShapes (anyway g_renderMode=0 and g_renderMode=1 are almost the same 90% of the times)
		//printf("g_renderMode=%d\n",g_renderMode);
	}
	GlutDemoApplicationEx::keyboardCallback(key,x,y);
}
void Hinge2VehicleDemo::specialKeyboardUp(int key, int x, int y){
	switch (key) {
   		case GLUT_KEY_LEFT :
    				g_vehicleKeyPressedLeft=false;
    				break;
    		case GLUT_KEY_RIGHT :
    				g_vehicleKeyPressedRight=false;
    				break;
    		case GLUT_KEY_UP :
    				g_vehicleKeyPressedUp=false;
    				break;
			case GLUT_KEY_DOWN :
					g_vehicleKeyPressedDown=false;
					break;
			case GLUT_KEY_F7: g_vehicleKeyPressedSpecialActionUp = false;
			case GLUT_KEY_F8: g_vehicleKeyPressedSpecialActionDown = false;
					// We need to look the vehicle 'specialAction' here:
					{
					IDrivingInterface* vehicle = m_activeVehicleIndex>=0 &&  m_activeVehicleIndex<m_driving_vehicles.size() ? m_driving_vehicles[m_activeVehicleIndex] : NULL;
					if (vehicle && vehicle->getChassis())	{
                        const int indexPlusOne = (const long int) GetUserPointer(vehicle->getChassis());
						if (indexPlusOne>0 && indexPlusOne>=g_sliderHelperStructs.size())	{
							const btSliderHelperStruct* shs = g_sliderHelperStructs[indexPlusOne-1];
							if (shs) shs->lock();
						}
					}
					}
					break;
            case GLUT_KEY_F4 : // Change camera target (i.e. drive another vehicle)
                    ++m_activeVehicleIndex;
					if (m_activeVehicleIndex>=m_driving_vehicles.size()) m_activeVehicleIndex=0;
                    refreshCameraVehicleOffset();
                    //printf("m_activeVehicleIndex=%d m_driving_vehicles.size()=%d m_minCameraDistance=%1.2f\n",m_activeVehicleIndex,m_driving_vehicles.size(),m_minCameraDistance);
                     break;
		default:
			GlutDemoApplicationEx::specialKeyboardUp(key,x,y);
        	break;
    }
}
void Hinge2VehicleDemo::specialKeyboard(int key, int x, int y){
    const int state = glutGetModifiers();   //This may issue a freeglut warning...

	IDrivingInterface* vehicle = m_activeVehicleIndex>=0 &&  m_activeVehicleIndex<m_driving_vehicles.size() ? m_driving_vehicles[m_activeVehicleIndex] : NULL;
	if (vehicle)
	{

		// SHIFT+ARROW KEYS = MOVE CAMERA
        if (state&BT_ACTIVE_SHIFT) {
			switch (key) 	{
    		case GLUT_KEY_LEFT :
    		case GLUT_KEY_RIGHT :
    		case GLUT_KEY_UP :
			case GLUT_KEY_DOWN :
				GlutDemoApplicationEx::specialKeyboard(key,x,y);	// Move camera
			return;
			}
		}

		// THESE KEYS MOVE VEHICLE (in preTickCallback())
		switch (key) 	{
    		case GLUT_KEY_LEFT :
    				g_vehicleKeyPressedLeft=true;
    				return;
    		case GLUT_KEY_RIGHT :
    				g_vehicleKeyPressedRight=true;
    				return;
    		case GLUT_KEY_UP :
                    {
    				//--Optional-things doesn't change so much-it's just a bit easier to get to a complete stop-----------------------
    				if (vehicle->isGoingBackwards()) vehicle->onBrake();	// when going backwards, we brake instead of accelerating
    				else
    				//-----------------------------------------------------------------------------------------------------------------
    					g_vehicleKeyPressedUp=true;
    				}
                    //else g_vehicleKeyPressedUp=true;
    				return;
			case GLUT_KEY_DOWN :
                    {
					//--Optional-things doesn't change so much-it's just a bit easier to get to a complete stop-----------------------
    				if (vehicle->isGoingForward()) vehicle->onBrake();		// when going forward, we brake instead of decelerating
    				else
    				//------------------------------------------------------------------------------------------------------------------
    					g_vehicleKeyPressedDown=true;
    				}
                    //else g_vehicleKeyPressedDown=true;
					return;
			case GLUT_KEY_F7:
					g_vehicleKeyPressedSpecialActionUp=true;
					return;
			case GLUT_KEY_F8:
					g_vehicleKeyPressedSpecialActionDown=true;
					return;
		}
	}

	// Other stuff not so useful:
	switch (key) 	{
		case GLUT_KEY_PAGE_DOWN :
			m_cameraDistance+=0.3f;
			//refreshCameraVehicleOffset();
			break;
		case GLUT_KEY_PAGE_UP :
			m_cameraDistance-=0.3f;if (m_cameraDistance<m_minCameraDistance) m_cameraDistance=m_minCameraDistance;
			//refreshCameraVehicleOffset();
			break;
        case GLUT_KEY_INSERT: break;
        case GLUT_KEY_HOME: break;
		default:
			GlutDemoApplicationEx::specialKeyboard(key,x,y);
        	break;
    }

//	glutPostRedisplay();


}
//endregion

 //region Camera Stuff
void Hinge2VehicleDemo::refreshCameraVehicleOffset()	{
	// This method just calculates m_minCameraDistance
	m_minCameraDistance = btScalar(1.f);
	IDrivingInterface* vehicle = m_activeVehicleIndex>=0 &&  m_activeVehicleIndex<m_driving_vehicles.size() ? m_driving_vehicles[m_activeVehicleIndex] : NULL;
	if (vehicle && vehicle->getChassis())	{
		const btRigidBody* chassis = vehicle->getChassis();
		const btVector3 aabb = GetAABBExtents(*chassis)*btScalar(0.5);
		m_minCameraDistance = m_minCameraDistance * aabb.length();
	}
}
void Hinge2VehicleDemo::refreshCameraRelatedVariables() {

	const int numVehicles = m_driving_vehicles.size();
	IDrivingInterface* vehicle = m_activeVehicleIndex>=0 &&  m_activeVehicleIndex<numVehicles ? m_driving_vehicles[m_activeVehicleIndex] : NULL;

	const btTransform chassisWorldTrans = vehicle ?
								GetGraphicWorldTransform(*vehicle->getChassis())
								: btTransform::getIdentity();

	//look at the vehicle
	m_cameraTargetPosition = chassisWorldTrans.getOrigin();
	if (m_cameraDistance<m_minCameraDistance) m_cameraDistance=m_minCameraDistance;

	const btMatrix3x3& cb = chassisWorldTrans.getBasis();

	const bool cameraOrientationFollowsChassisYPR = false;	// If true the camera tilts with the vehicle (not what it's naturally expected: good for planes)
	if (cameraOrientationFollowsChassisYPR)	{
		m_cameraT.getBasis()=cb;
		m_cameraT.getBasis()*=btMatrix3x3(btQuaternion(btVector3(0,1,0),btRadians(m_azi))*btQuaternion(btVector3(1,0,0),btRadians(m_ele)));
	}
	else	{
			const bool cameraOrientationFollowsOnlyChassisYaw = true;//false;	// If true the camera pitch is absolute
															//m_cameraDistance > m_minCameraDistance*3;	// true makes me feel sick when the camera is too close...

			btScalar cbAziRad=0,cbEleRad=0;
			btMatrix3x3Helper::ToEulerYXZ(cb,cbAziRad,cbEleRad);
			if (cameraOrientationFollowsOnlyChassisYaw) cbEleRad = 0;

			m_cameraT.getBasis()=btMatrix3x3(btQuaternion(btVector3(0,1,0),btRadians(m_azi)+cbAziRad)*btQuaternion(btVector3(1,0,0),btRadians(m_ele)+cbEleRad));
	}

	m_cameraT.getOrigin() = m_cameraPosition = m_cameraTargetPosition - m_cameraT.getBasis().getColumn(2)*m_cameraDistance;
	m_cameraUp.setValue(0,1,0);

}
//endregion


btHinge2Vehicle* Hinge2VehicleDemo::localCreateHinge2Vehicle(
											 const btTransform& startTransform,
											 const unsigned short numberOfWheels,
											 const unsigned short numberOfSteeringWheels,
											 const unsigned short tractionMode,			// 0 = steering wheels, 1 = not steering wheels, 2 = all wheels ( NO DIFFERENTIAL IN ALL CASES )
											 const unsigned short wheelShapeTypeIndex,	// 0 = cyl, 1 = scaled sphere, 2 = very complex wheel shape, 3 = train wheel shape
											 const unsigned short wheelPositionMode,	// 0 = NORMAL, 1 = PUT ALL WHEELS AT THE FRONT, 2 = PUT ALL WHEELS AT THE BACK


											 const btTriangleMeshEx& chassisCenteredMesh,
											 const btTriangleMeshEx& wheelCenteredMesh,

											 const btHACDCompoundShape::Params& chassisHACDCompoundShapeParams,
											 const bool useFlatNormalsForChassisMesh,

											 const btScalar _chassisMass,
											 const btScalar _wheelMass,

											 const btVector3& chassisCOM,
											 const btVector3& _chassisWheelOffset,
											 const btVector3& wheelsAsymmetricOffset,

											 const btScalar chassisFriction,
											 const btScalar chassisRestitution,
											 const btScalar wheelFriction,
											 const btScalar wheelRestitution,

											 btCollisionShape* reuseThisCollisionShapeForChassis,	//If used, all other params must be set as if it was NULL
											 btCollisionShape* reuseThisCollisionShapeForWheel,		//If used, all other params must be set as if it was NULL

											 const short int chassisCollisionFilterGroup,
											 const short int chassisCollisionFilterMask,
											 const short int wheelsCollisionFilterGroup,
											 const short int wheelsCollisionFilterMask,

											 // These should be left as default unless further tweaking is required...
											 const btScalar verticalSlidersExtension,
											 const btScalar verticalSlidersMotorVelocity,
											 const btScalar verticalSlidersMotorMaxForce
											)	{
	//====================== PARAMS CHECKING ====================================
	const btVector3 chassisLocalAABB(chassisCenteredMesh.getAabbFullExtents());
	btVector3 COM(chassisCOM);//if (COM==btVector3(90.f,0.f,0.f)) COM=btVector3(0.f,-chassisLocalAABB.y()*0.5f,0.f);
	btScalar chassisMass(_chassisMass);if (chassisMass<=0) chassisMass = 100.0f*chassisLocalAABB.x()+chassisLocalAABB.y()+chassisLocalAABB.z();

	const btVector3 wheelLocalAABB(wheelCenteredMesh.getAabbFullExtents());
	const btScalar wheelRadius(wheelLocalAABB.z()*0.5f);
	const btScalar wheelWidth(wheelLocalAABB.x());
	btScalar wheelMass(_wheelMass); if(wheelMass<=0) wheelMass=(wheelLocalAABB.y()*SIMD_PI*wheelLocalAABB.x()*20.0f);

	btVector3 chassisWheelOffset(_chassisWheelOffset); if (chassisWheelOffset==btVector3(0.f,0.f,0.f))	chassisWheelOffset = btVector3(chassisLocalAABB.x()*0.5f-wheelWidth*0.25f,-chassisLocalAABB.y()*0.5f-wheelRadius*0.5f,chassisLocalAABB.z()*0.5f-wheelRadius);
	//===========================================================================

	btCollisionShape* chassisShape = NULL;
	if (reuseThisCollisionShapeForChassis) chassisShape = reuseThisCollisionShapeForChassis;
	else	{
		btCompoundShape* _chassisShape = new btHACDCompoundShape(const_cast < btTriangleMeshEx& > (chassisCenteredMesh),chassisHACDCompoundShapeParams);
		// Here we add the COM offset directly inside the collision shape:
		for (int i=0,sz=_chassisShape->getNumChildShapes();i<sz;i++)	{
			btTransform& T(_chassisShape->getChildTransform(i));
			T.setOrigin(T.getOrigin()- COM*T.getBasis() ); // Not sure if it's correct, but here T.getBasis()==identity...
			_chassisShape->updateChildTransform(i,T,(i==sz-1));
		}
		chassisShape = _chassisShape;
		AssignMesh(chassisShape,chassisCenteredMesh,useFlatNormalsForChassisMesh);
		// What have we done! We have just assigned chassisShape ( with COM offset embedded) to chassisCenteredMesh ( no COM offset ).
		// This is not a smart move at all! However to overcome this we use 2 different ways:
		// 1) If USE_COLLISION_OBJECT_USER_POINTER is defined (in "GlutDemoApplicationEx.h") we just add a COM offset in the motion state (see below).
		// 2) If USE_COLLISION_OBJECT_USER_POINTER is not defined (in "GlutDemoApplicationEx.h"), we just add the graphic offset as a mesh override.
		// The first solution has the advantage that the transform OUTSIDE the COM can be retrieved through the motion state (very good for the camera)

		// Hack (but usually USE_COLLISION_OBJECT_USER_POINTER is defined (in "GlutDemoApplicationEx.h")):
		#ifndef USE_COLLISION_OBJECT_USER_POINTER
		if (chassisShape) GetMeshOverrides(*chassisShape).offset = -COM;
		#endif //USE_COLLISION_OBJECT_USER_POINTER
	}
	 //region Build Body
	btVector3 localInertia(0,0,0);
	chassisShape->calculateLocalInertia(chassisMass,localInertia);
	const bool useCOMoffsetIndeed(COM!=btVector3(0,0,0));
	//using motionstate is recommended, it provides interpolation capabilities, and only synchronizes 'active' objects
	btDefaultMotionState* myMotionState;
	if (!useCOMoffsetIndeed)	myMotionState = new btDefaultMotionState(startTransform);
	else	{
		#ifndef USE_COLLISION_OBJECT_USER_POINTER
		const btTransform transformedStartTransform(startTransform.getBasis(),startTransform.getOrigin()+startTransform.getBasis()*COM);
		myMotionState = new btDefaultMotionState(transformedStartTransform);
		#else //USE_COLLISION_OBJECT_USER_POINTER
		myMotionState = new btMyDefaultMotionState(startTransform,COM);
		#endif//USE_COLLISION_OBJECT_USER_POINTER
	}
	btRigidBody::btRigidBodyConstructionInfo cInfo(chassisMass,myMotionState,chassisShape,localInertia);
	btRigidBody* chassis = new btRigidBody(cInfo);
	if (chassis)	{
		chassis->setContactProcessingThreshold(m_defaultContactProcessingThreshold);
		const bool isDynamic(true);
		const short int _chassisCollisionFilterGroup = chassisCollisionFilterGroup>=0 ? (isDynamic ? chassisCollisionFilterGroup : (chassisCollisionFilterGroup|btBroadphaseProxy::StaticFilter)) : (isDynamic? short(btBroadphaseProxy::DefaultFilter) : short(btBroadphaseProxy::StaticFilter));
		const short int _chassisCollisionFilterMask = chassisCollisionFilterMask>=0 ? (isDynamic ? chassisCollisionFilterMask : (chassisCollisionFilterMask&(~btBroadphaseProxy::StaticFilter))) : (isDynamic? short(btBroadphaseProxy::AllFilter) : short(btBroadphaseProxy::AllFilter ^ btBroadphaseProxy::StaticFilter));
		m_dynamicsWorld->addRigidBody(chassis,_chassisCollisionFilterGroup,_chassisCollisionFilterMask);
	}
	//endregion
	if (chassis)	{
		chassis->setFriction(chassisFriction);
		chassis->setRestitution(chassisRestitution);
	}

	btCollisionShape* wheelShape = NULL;
	if (reuseThisCollisionShapeForWheel) wheelShape = reuseThisCollisionShapeForWheel;
	else	{
		switch (wheelShapeTypeIndex%4) {
			case 0: wheelShape = new btWheelCylinderShape(wheelRadius,wheelWidth);break;
			case 1: wheelShape = new btWheelSimpleShape(wheelRadius,wheelWidth);break;
			case 2: wheelShape = new btWheelShape(wheelRadius,wheelWidth,16,0.95);break;		// This gets a bit bigger than needed!
			case 3: wheelShape = new btWheelTrainShape(wheelRadius,wheelWidth);break;
		}
		//m_collisionShapes.push_back(wheelShape);
		//#define DEBUG_ME
		// Sometimes the visual appearance of the wheel shape seems of different sizes,
		// depending on the wheelShapeTypeIndex. So we can debug the (approximate?) real size here:
		#ifdef DEBUG_ME
			btVector3 aabbMin,aabbMax,aabbExtents;
			wheelShape->getAabb(btTransform::getIdentity(),aabbMin,aabbMax);
			aabbExtents = (aabbMax - aabbMin);
			printf("Wheel shape local AABB: (%10.2f,%10.2f,%10.2f)\n",aabbExtents.x(),aabbExtents.y(),aabbExtents.z());
			printf("Wheel shape width:		%10.2f\n",aabbExtents.x());
			printf("Wheel shape radius:		%10.2f\n",0.5f*(aabbExtents.y()+aabbExtents.z())*0.5f);
			#undef DEBUG_ME
		#endif //DEBUG_ME

			if (wheelShape) AssignMesh(wheelShape,wheelCenteredMesh);

	}


	//const int wheelRows = numberOfWheels==2 ? 2 : numberOfWheels/2 + numberOfWheels%2;
	const bool firstWheelAtCenter = numberOfWheels==2 || numberOfWheels%2 ? true : false;
	//const bool lastWheelAtCenter = numberOfWheels==2 ? true : false;	// Only true for 2 wheels for now
	btAlignedObjectArray<btRigidBody*> wheels;
	const unsigned short positionMode = wheelPositionMode>3 ? 0 : wheelPositionMode;
	const bool putAllWheelsForward = positionMode==1;
	const bool putAllWheelsBackward = positionMode==2;
	const bool putSteeringWheelsAtTheBack = positionMode==3;
	btRigidBody* wheel;	// temp
	btTransform T;	//temp;
	{
	 //region Some base transforms to start with
	const btMatrix3x3 rot180(btQuaternion(btVector3(0.f,1.f,0.f),SIMD_PI));
	// center (below vehicle)
	btTransform T_y(startTransform);
	T_y.setOrigin(T_y.getOrigin()+wheelsAsymmetricOffset);
	T_y.setOrigin(T_y.getOrigin()+T_y.getBasis().getColumn(1)*(/*-COM.y()*/+chassisWheelOffset.y()));
	// forward center
	btTransform T_y_Zf(T_y);
	T_y_Zf.setOrigin(T_y_Zf.getOrigin()+T_y_Zf.getBasis().getColumn(2)*(/*-COM.z()*/+chassisWheelOffset.z()));
	// backward center
	btTransform T_y_Zb(T_y);
	T_y_Zb.setOrigin(T_y_Zb.getOrigin()+T_y_Zb.getBasis().getColumn(2)*(/*-COM.z()*/-chassisWheelOffset.z()));
	// forward left edge
	btTransform T_y_Zf_Xl(T_y_Zf);
	T_y_Zf_Xl.setOrigin(T_y_Zf_Xl.getOrigin()+T_y_Zf_Xl.getBasis().getColumn(0)*(/*-COM.x()*/+chassisWheelOffset.x()));
	// forward right edge
	btTransform T_y_Zf_Xr(T_y_Zf);
	T_y_Zf_Xr.setOrigin(T_y_Zf_Xr.getOrigin()+T_y_Zf_Xr.getBasis().getColumn(0)*(/*-COM.x()*/-chassisWheelOffset.x()));	// MMMhh, we should have further rotated it 180 degres around the Y axis...
	btTransform T_y_Zf_Xr_180(T_y_Zf_Xr);T_y_Zf_Xr_180.setBasis(T_y_Zf_Xr_180.getBasis()*rot180);
	// backward left edge
	btTransform T_y_Zb_Xl(T_y_Zb);
	T_y_Zb_Xl.setOrigin(T_y_Zb_Xl.getOrigin()+T_y_Zb_Xl.getBasis().getColumn(0)*(/*-COM.x()*/+chassisWheelOffset.x()));
	// backward right edge
	btTransform T_y_Zb_Xr(T_y_Zb);
	T_y_Zb_Xr.setOrigin(T_y_Zb_Xr.getOrigin()+T_y_Zb_Xr.getBasis().getColumn(0)*(/*-COM.x()*/-chassisWheelOffset.x()));	// MMMhh, we should have further rotated it 180 degres around the Y axis...
	btTransform T_y_Zb_Xr_180(T_y_Zb_Xr);T_y_Zb_Xr_180.setBasis(T_y_Zb_Xr_180.getBasis()*rot180);
	//endregion
	if (numberOfWheels==1 && positionMode == 0)	{
		T = positionMode==1 ? T_y_Zf : positionMode==2 ? T_y_Zb : T_y;
		wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
	}
	else if (numberOfWheels==2 && positionMode == 0)	{
		T = positionMode==1 ? T_y_Zf_Xl : positionMode==2 ? T_y_Zb_Xl : T_y_Zf;
		wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
		T = positionMode==1 ? T_y_Zf_Xr : positionMode==2 ? T_y_Zb_Xr : T_y_Zb;
		wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
	}
	else if (positionMode == 0)	{
		// Front Wheels
		{
			//TODO: in positionMode=0, we can improve e.g. 8 wheels, by putting 4 in front and 4 in back (now 2 in front and 6 in back)
			if (firstWheelAtCenter) {
				wheel = localCreateRigidBody(wheelMass,T_y_Zf,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
			}
			else {
				wheel = localCreateRigidBody(wheelMass,T_y_Zf_Xl,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				wheel = localCreateRigidBody(wheelMass,T_y_Zf_Xr_180,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
			}
		}
		// Rear Wheels
		{
			const int numRowsLeft = (numberOfWheels - wheels.size())/2;// + lastWheelAtCenter ? 1 : 0;
			for (int row = numRowsLeft-1; row >= 0;row--)	{
				T=T_y_Zb_Xl;
				T.setOrigin(T.getOrigin()+T.getBasis().getColumn(2)*(2.f*wheelRadius+wheelRadius*0.25f)*(row/*+lastWheelAtCenter ? 1 : 0*/));
				wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				T=T_y_Zb_Xr;
				T.setOrigin(T.getOrigin()+T.getBasis().getColumn(2)*(2.f*wheelRadius+wheelRadius*0.25f)*(row/*+lastWheelAtCenter ? 1 : 0*/));
				T.setBasis(T.getBasis()*rot180);
				wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				}
			/*
			if (lastWheelAtCenter) {
				wheel = localCreateRigidBody(wheelMass,T_y_Zb,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
			}
			*/
		}
	}
	else if (positionMode == 1)	{
		// Front Wheels
		{
			if (firstWheelAtCenter) {
				wheel = localCreateRigidBody(wheelMass,T_y_Zf,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
			}
			const int numRowsLeft = numberOfWheels/2;
			for (int row = numRowsLeft-1; row >= 0;row--)	{
				T=T_y_Zf_Xl;
				T.setOrigin(T.getOrigin()+T.getBasis().getColumn(2)*(2.f*wheelRadius+wheelRadius*0.25f)*(row+firstWheelAtCenter ? 1 : 0));
				wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				T=T_y_Zf_Xr;
				T.setOrigin(T.getOrigin()+T.getBasis().getColumn(2)*(2.f*wheelRadius+wheelRadius*0.25f)*(row+firstWheelAtCenter ? 1 : 0));
				T.setBasis(T.getBasis()*rot180);
				wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				}
		}
	}
	else if (positionMode == 2)	{
		// Rear Wheels
		{
			const bool lastWheelAtCenter = (numberOfWheels%2)==1;
			const int numRowsLeft = numberOfWheels/2;
			for (int row = numRowsLeft-1; row >= 0;row--)	{
				T=T_y_Zb_Xl;
				T.setOrigin(T.getOrigin()+T.getBasis().getColumn(2)*(2.f*wheelRadius+wheelRadius*0.25f)*(row+lastWheelAtCenter ? 1 : 0));
				wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				T=T_y_Zb_Xr;
				T.setOrigin(T.getOrigin()+T.getBasis().getColumn(2)*(2.f*wheelRadius+wheelRadius*0.25f)*(row+lastWheelAtCenter ? 1 : 0));
				T.setBasis(T.getBasis()*rot180);
				wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				}
			if (lastWheelAtCenter) {
				wheel = localCreateRigidBody(wheelMass,T_y_Zb,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
			}

		}
	}
	else if (positionMode == 3)	{
		// Front Wheels
		{
			if (firstWheelAtCenter) {
				wheel = localCreateRigidBody(wheelMass,T_y_Zb,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
			}
			else {
				wheel = localCreateRigidBody(wheelMass,T_y_Zb_Xl,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				wheel = localCreateRigidBody(wheelMass,T_y_Zb_Xr_180,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
			}
		}
		// Rear Wheels
		{
			const int numRowsLeft = (numberOfWheels - wheels.size())/2;// + lastWheelAtCenter ? 1 : 0;
			for (int row = numRowsLeft-1; row >= 0;row--)	{
				T=T_y_Zf_Xl;
				T.setOrigin(T.getOrigin()+T.getBasis().getColumn(2)*(2.f*wheelRadius+wheelRadius*0.25f)*(row/*+lastWheelAtCenter ? 1 : 0*/));
				wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				T=T_y_Zf_Xr;
				T.setOrigin(T.getOrigin()+T.getBasis().getColumn(2)*(2.f*wheelRadius+wheelRadius*0.25f)*(row/*+lastWheelAtCenter ? 1 : 0*/));
				T.setBasis(T.getBasis()*rot180);
				wheel = localCreateRigidBody(wheelMass,T,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
				}
			/*
			if (lastWheelAtCenter) {
				wheel = localCreateRigidBody(wheelMass,T_y_Zf,wheelShape,wheelsCollisionFilterGroup,wheelsCollisionFilterMask);if (wheel) wheels.push_back(wheel);
			}
			*/
		}
	}

	}

	for (int i=0;i<wheels.size();i++)	{
		wheel = wheels[i];
		if (wheel)	{
			wheel->setFriction(wheelFriction);
			wheel->setRestitution(wheelRestitution);
            //wheel->setRollingFriction(0);
            //wheel->setCollisionFlags(wheel->getCollisionFlags()|CF_HIDE_SHADOW);	// No stencil shadow for it, to speed up code a bit
			#ifdef USE_ANISOTROPIC_FRICTION_FOR_WHEELS
			// I don't know if this works, but it seems to work: if you raise the friction.x component vehicles fall on their side
			// when steering (the lack of differential increase this effect):
			if (wheelShapeTypeIndex%4!=3) wheel->setAnisotropicFriction(btVector3(g_xAnisotropicFrictionFactorForWheels*wheelFriction,wheelFriction,wheelFriction));
			#endif //USE_ANISOTROPIC_FRICTION_FOR_WHEELS
            #ifdef BETTER_STABILITY_OF_WHEEL_TRAIN_SHAPES
            // Train wheel only:
            if (wheelShapeTypeIndex%4==3) {
                #if BETTER_STABILITY_OF_WHEEL_TRAIN_SHAPES==2
                if (tractionMode!=0)
                #endif
                wheel->setCollisionFlags(wheel->getCollisionFlags()  | btCollisionObject::CF_CUSTOM_MATERIAL_CALLBACK);
            }
            #endif //BETTER_STABILITY_OF_WHEEL_TRAIN_SHAPES
        }
	}

    btHinge2Vehicle* vehicle = new btHinge2Vehicle(m_dynamicsWorld,chassis,wheels,numberOfSteeringWheels<0 ? 0 : numberOfSteeringWheels>wheels.size() ? wheels.size() : numberOfSteeringWheels,tractionMode,verticalSlidersExtension,verticalSlidersMotorVelocity,verticalSlidersMotorMaxForce);
        if (vehicle) {
            m_hinge2Vehicles.push_back(vehicle);//garbage collector
            if (positionMode==3) vehicle->setInvertSteeringKeys(true);
            vehicle->setSteeringReturnConstant(vehicle->getSteeringReturnConstant()*wheelRadius*4);	// Naive? Well, since the motor velocity is angular, the return constant should be greater with the wheel perimeter (or not?)
        }

	return vehicle;
}



