// Based on Bullet ConcavePhysicsDemo.
/*
Bullet Continuous Collision Detection and Physics Library
Copyright (c) 2003-2006 Erwin Coumans  http://continuousphysics.com/Bullet/

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/
#pragma once
#ifndef HINGE2_VEHICLE_DEMO_H
#define HINGE2_VEHICLE_DEMO_H

class btCollisionShape;

#include "GlutDemoApplicationEx.h"
#include "Optionals/btHACDCompoundShape.h"
#include "btHinge2Vehicle.h"



class Hinge2VehicleDemo : public GlutDemoApplicationEx
{
	public:

	btScalar m_minCameraDistance;
	void refreshCameraVehicleOffset();

	btAlignedObjectArray< btHinge2Vehicle* > m_hinge2Vehicles;	// All the vehicle parts. For garbage collection
	
	btAlignedObjectArray<IDrivingInterface* > m_driving_vehicles;	// A subset of the above, allowing driving input and camera targeting, casted to their driving interfaces (I could have simply used "btAlignedObjectArray<btHinge2Vehicle* >" as above, I did it just to confuse newbies...) 
	int m_activeVehicleIndex;	// Index of the above array

	Hinge2VehicleDemo();

	virtual ~Hinge2VehicleDemo()	{
		exitPhysics();
	}

	virtual void createEmptyDynamicsWorld(bool softBodySupport = false);
	
	virtual void createGround();
	
	virtual void clientMoveAndDisplay();
	
	///a very basic camera following the vehicle
	virtual void refreshCameraRelatedVariables();

	virtual void specialKeyboard(int key, int x, int y);

	virtual void specialKeyboardUp(int key, int x, int y);

	void initPhysics();

    void initPhysicsHinge2Vehicles();
	
	void exitPhysics();
	
	void setVertexPositions(float waveheight, float offset);
	
	void keyboardCallback(unsigned char key, int x, int y);
	
	virtual void preTickCallback(btScalar timeStep);

	// Helper method that contain collision shape creation and creation of btHinge2Vehicles
	btHinge2Vehicle* localCreateHinge2Vehicle(
											 const btTransform& startTransform,
											 const unsigned short numberOfWheels,
											 const unsigned short numberOfSteeringWheels,
											 const unsigned short tractionMode,			// 0 = steering wheels, 1 = not steering wheels, 2 = all wheels ( NO DIFFERENTIAL IN ALL CASES )
											 const unsigned short wheelShapeTypeIndex,	// 0 = cyl, 1 = scaled sphere, 2 = very complex wheel shape, 3 = train wheel shape
											 												// WARNING AT THE MOMENT OF WRITING: AFAIK the Bullet OpenGL support does not support local scaled shapes (at least btMultiSphereShapes):
											 												// this results in wrong and independent wheelWidth scaling when using shape types 2 and 3 (this only affects graphics).
											 const unsigned short wheelPositionMode,	// 0 = NORMAL, 1 = PUT ALL WHEELS AT THE FRONT, 2 = PUT ALL WHEELS AT THE BACK
											 
											 const btTriangleMeshEx& chassisCenteredMesh,
											 const btTriangleMeshEx& wheelCenteredMesh,	
											 
											 const btHACDCompoundShape::Params& chassisHACDCompoundShapeParams=btHACDCompoundShape::Params(),
											 const bool useFlatNormalsForChassisMesh= false,										 											 
											 
											 const btScalar _chassisMass=btScalar(-1.f),
											 const btScalar _wheelMass=btScalar(-1),

											 const btVector3& chassisCOM=btVector3(0.f,0.f,0.f),
											 const btVector3& _chassisWheelOffset=btVector3(0.f,0.f,0.f),		// position of the front-right wheel relative to the chassis mesh center
											 const btVector3& _wheelsAsymmetricOffset=btVector3(0.f,0.f,0.f),	// mainly useful to shift all the wheels together to the front/back

											 const btScalar chassisFriction=btScalar(0.5f),
											 const btScalar chassisRestitution=btScalar(0.0f),
											 const btScalar wheelFriction=btScalar(1.0f),
											 const btScalar wheelRestitution=btScalar(0.0f),
											 
											 btCollisionShape* reuseThisCollisionShapeForChassis = NULL,	//If used, all other params must be set as if it was NULL. Use it only with chassis shapes previously created by this same method.
											 btCollisionShape* reuseThisCollisionShapeForWheel = NULL,		//If used, all other params must be set as if it was NULL. Use it only with wheel shapes previously created by this same method.											 
											 
											 const short int chassisCollisionFilterGroup=-1,
											 const short int chassisCollisionFilterMask=-1,
											 const short int wheelsCollisionFilterGroup=-1,
											 const short int wheelsCollisionFilterMask=-1,											 
											 
											 // These should be left as default unless further tweaking is required...
											 const btScalar verticalSlidersExtension=btScalar(-1),
											 const btScalar verticalSlidersMotorVelocity=btScalar(-1),
											 const btScalar verticalSlidersMotorMaxForce=btScalar(-1)												 
											);		
	

};

#endif //HINGE2_VEHICLE_DEMO_H


